/*
 * QUSOFT MICROSYSTMES
 * Power Basic
 * Copyright 2004 Frdric Brown
 */

/*
 *  This file is part of Power Basic library.
 *
 *  Power Basic library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Power Basic library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Power Basic library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* This file implements the Variant type */

#include "Variant.h"

char getFontWidth() {
	switch (FontGetSys()) {
		case F_4x6 :
			return 4;
			break;
		case F_6x8 :
			return 6;
			break;
		default :
			return 8;
			break;
	}
}

char getFontHeight() {
	switch (FontGetSys()) {
		case F_4x6 :
			return 6;
			break;
		case F_6x8 :
			return 8;
			break;
		default :
			return 10;
			break;
	}
}

Variant newVariantArray(VariantArray * array) {
	Variant v;

	v.type = VARIANT_ARRAY;

	v.array = array;

	return v;
}

Variant  newVariantStr(char * str) {
	Variant v;

	v.type = VARIANT_STRING;

	strcpy(v.str, str);

	return v;
}

Variant newVariantNum(double num) {
	Variant v;

	v.type = VARIANT_NUMERIC;

	v.num = num;

	return v;
}

Variant newVariantPtr(Variant * ptr) {
	Variant v;

	v.type = VARIANT_POINTER;

	v.ptr = ptr;

	return v;
}

Variant asVariantString(Variant v) {
	Variant n;
	n.type = VARIANT_STRING;

	switch (v.type) {
		case VARIANT_NUMERIC: {
			sprintf(n.str, "%f", v.num);
			break;
		}
		case VARIANT_STRING: {
			strcpy(n.str, v.str);
			break;
		}
		case VARIANT_POINTER: {
			sprintf(n.str, "%lp", v.ptr);
			break;
		}
		default : {
			n.str[0] = '\0';
		}
	}

	return n;
}

double strToFloat(char * str) {
	char buffer[VARIANT_STRING_SIZE];
	strcpy(buffer, str);
	short i;
	char c;

	for (i = 0; (c = buffer[i]); i++)
	{
		if (c == '-') buffer[i] = 0xAD;
		if ((c|32) == 'e') buffer[i] = 0x95;
	}

	return atof(buffer);
}

Variant asVariantNum(Variant v) {
	Variant n;

	n.type = VARIANT_NUMERIC;

	switch (v.type) {
		case VARIANT_NUMERIC: {
			n.num = v.num;
			break;
		}
		case VARIANT_STRING: {

			n.num = strToFloat(v.str);

			break;
		}
		default : {
			n.type = VARIANT_UNDEFINED;
		}
	}

	return n;
}

Variant asVariantPtr(Variant v) {
	Variant n;
	n.type = VARIANT_POINTER;

	switch (v.type) {
		case VARIANT_STRING: {
			n.ptr = (Variant *) atol(v.str);
			break;
		}
		case VARIANT_POINTER: {
			n.ptr = v.ptr;
			break;
		}
		default : {
			n.type = VARIANT_UNDEFINED;
		}
	}

	return n;
}

double variantCompareTo(Variant v1, Variant v2) {
	switch (v1.type) {
		case VARIANT_NUMERIC: {
			return (v1.num - v2.num);
		}
		case VARIANT_STRING: {
			return (strcmp(v1.str, v2.str));
		}
		case VARIANT_POINTER: {
			return (v1.ptr - v2.ptr);
		}
		default: {
			return -1;
		}
	}
}

Variant concatVariantStr(Variant v1, Variant v2) {

	strcat(v1.str, v2.str);

	return v1;
}

ESI Variant_argPtr; //Initialiser  NULL

Variant fetchVariantFromArg() {
	Variant v;

	if (!Variant_argPtr)
	{
		InitArgPtr(Variant_argPtr);
	}

	v.type = VARIANT_NUMERIC;

	switch (GetArgType(Variant_argPtr)) {
		case POSINT_TAG: {
			v.num = GetIntArg(Variant_argPtr);
			break;
		}
		case NEGINT_TAG: {
			v.num = -1 * GetIntArg(Variant_argPtr);
			break;
		}
		case FLOAT_TAG: {
			v.num = GetFloatArg(Variant_argPtr);
			break;
		}
		case POSFRAC_TAG: {
			v.num = GetIntArg(Variant_argPtr);
			Variant_argPtr++;
			v.num /= GetIntArg(Variant_argPtr);
			break;
		}
		case NEGFRAC_TAG: {
			v.num = -1 * GetIntArg(Variant_argPtr);
			Variant_argPtr++;
			v.num /= GetIntArg(Variant_argPtr);
			break;
		}
		case STR_TAG: {
			v.type = VARIANT_STRING;
			v = newVariantStr((char*)GetStrnArg(Variant_argPtr));
			break;
		}
		default: {
			ER_throw (ER_TOO_FEW_ARGS);
			v.type = VARIANT_UNDEFINED;
			v.num = 0;
		}
	}

	return v;
}

void command_Exec (char * str) {
	char buffer[VARIANT_STRING_SIZE + 7];

	strcpy(buffer, "Exec \"");
	strcat(buffer, str);
	strcpy(buffer, "\"");

	execString(buffer);
}

void execString (char * str) {
		HANDLE handle;

		push_parse_text (str);
		handle = HS_popEStack ();

		TRY
    		NG_execute (handle, FALSE);
  		FINALLY
    		HeapFree (handle);
		ENDFINAL
}

void exportVariant(Variant *v, char * varName) {
	char buffer[256];
	Variant vs = asVariantString(*v);
	short len = strlen(vs.str);

	strcpy(buffer, vs.str);
	buffer[len] = (char)22;
	buffer[len+1] = '\0';
	strcat(buffer, varName);

	execString(buffer);
}

void importVariant(Variant *v, char * varName) {
	SYM_ENTRY *entry;
	char buffer[256];
	ESI Variant_argPtr;
	BOOL locked;

	entry = DerefSym(SymFind(toVATBuffer(buffer, varName)));

	locked = HeapGetLock(entry->handle);

	if (!locked) {
		HeapLock(entry->handle);
	}

	Variant_argPtr = HToESI(entry->handle);

	v->type = VARIANT_NUMERIC;

	switch (GetArgType(Variant_argPtr)) {
		case POSINT_TAG: {
			v->num = GetIntArg(Variant_argPtr);
			break;
		}
		case NEGINT_TAG: {
			v->num = -1 * GetIntArg(Variant_argPtr);
			break;
		}
		case FLOAT_TAG: {
			v->num = GetFloatArg(Variant_argPtr);
			break;
		}
		case POSFRAC_TAG: {
			v->num = GetIntArg(Variant_argPtr);
			Variant_argPtr++;
			v->num /= GetIntArg(Variant_argPtr);
			break;
		}
		case NEGFRAC_TAG: {
			v->num = -1 * GetIntArg(Variant_argPtr);
			Variant_argPtr++;
			v->num /= GetIntArg(Variant_argPtr);
			break;
		}
		case STR_TAG: {
			v->type = VARIANT_STRING;
			strcpy(v->str, (char*)GetStrnArg(Variant_argPtr));
			break;
		}
		default: {
			ER_throw (ER_TOO_FEW_ARGS);
			v->type = VARIANT_UNDEFINED;
			v->num = 0;
		}
	}

	if (!locked) {
		HeapUnlock(entry->handle);
	}
}

void pushVariant(Variant *v) {
	switch (v->type) {
		case VARIANT_STRING: {
			char* output_string = malloc(strlen(v->str) + 2);
			//put the zero at the beginning of the string
			output_string[0] = 0;
			//copy the string, including the terminating zero, to the output string
			strcpy(output_string + 1, v->str);

			//put the string, with all spaces intact, on the expression stack
			push_expr_quantum(output_string + strlen(output_string + 1) + 1, STR_TAG);

			//always make sure to free your memory
			free(output_string);
			break;
		}
		case VARIANT_POINTER: {
			pushVariant(v->ptr);
			break;
		}
		case VARIANT_ARRAY: {
			pushVariant(v->ptr);
			break;
		}
		default : {
			push_Float(v->num);
		}
	}
}

void returnVariant(Variant v) {
	TRY
		while (GetArgType (top_estack) != END_TAG)
	  		top_estack = next_expression_index (top_estack);
		top_estack--;
	ONERR
	ENDTRY


	pushVariant(& v);

	exit(0);
}

void initVariantType() {
	Variant_argPtr = NULL;
}

void printVariant(Variant v) {
	short i;
	short n;
	
	switch (v.type) {
		case VARIANT_NUMERIC: {
			printf("%f", v.num);
			break;
		}
		case VARIANT_STRING: {
			printf(v.str);
			break;
		}
		case VARIANT_POINTER: {
			printf("%lp", v.ptr);
			break;
		}
		case VARIANT_ARRAY: {
			for (i = 0; i < v.array->rNum; i++) {
				printf("[");
					for (n = 0; n < v.array->cNum; n++) {
						if (n > 0) {
							printf(", ");
						}
						printVariant(v.array->ptr[i][n]);
					}
				printf("]");
			}
			break;
		}
		default : {
			/*nop*/
		}
	}
}

char * toVATBuffer(char * buffer, char * str) {
	buffer[0] = '\0';
	strcpy(&buffer[1], str);

	return &buffer[strlen(&buffer[1]) + 1];
}

void newArray(VariantArray * a, short r, short c) {
	short i;
	
	a->rNum = r;
	a->cNum = c;
	
	a->ptr = (Variant **) malloc (sizeof(Variant*)*r);
	
	for (i = 0; i < r; i++) {
		a->ptr[i] = (Variant *)malloc (sizeof(Variant)*c);
	}
}

void redimArray(VariantArray * a, short r, short c) {
	short i;
	
	if (a->rNum != r) {
		for (i = a->rNum-1; i > r - 1; i--) {
			free ( a->ptr[i] );
		}
		a->ptr = (Variant **) realloc(a->ptr, sizeof(Variant*)*r);
		for (i = a->rNum-1; i < r; i++) {
			a->ptr[i] = (Variant *)malloc (sizeof(Variant)*c);
		}
		a->rNum = r;
	}
	
	for (i = 0; i < a->rNum; i++) {
		a->ptr[i] = (Variant *)realloc (a->ptr[i], sizeof(Variant)*c);
	}
	
	a->cNum = c;
}

void freeArray(VariantArray * a) {
	short i;
	
	for (i = a->rNum-1; i >= 0; i--) {
		free ( a->ptr[i] );
	}
	
	free ( a->ptr );
}

void getInput(Variant *v) {
	char buffer[VARIANT_STRING_SIZE];
	
	getsn (buffer, VARIANT_STRING_SIZE);
	
	if ((strlen(buffer) < 1) || (buffer[0] != '"')) {	
		*v = newVariantNum(strToFloat(buffer));
	}
	else {
		buffer[strlen(buffer)-1] = '\0';
		*v = newVariantStr((buffer + 1));
	}
}

short convertAttr (short o) {
	
	switch(o) {
		case 1 :
			return A_NORMAL;
		case 0 :
			return A_REVERSE;
		case -1 :
			return A_XOR;
	}
	
	return 1;
}
