(*
 * QUSOFT MICROSYSTMES
 * Power Basic
 * Copyright 2004 Frdric Brown
 *)

(*
 *  This file is part of Power Basic.
 *
 *  Power Basic is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.

 *  Power Basic is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with Power Basic; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit BuildFrame;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls;

type
  TFormBuild = class(TForm)
    Label1: TLabel;
    txtDir: TEdit;
    Label2: TLabel;
    btnConvert: TButton;
    btnCancel: TButton;
    Label3: TLabel;
    txtMes: TMemo;
    pnlMes: TPanel;
    btnClose: TButton;
    procedure btnCancelClick(Sender: TObject);
    procedure btnConvertClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

 // function commence(s: AnsiString; vec: TStrings): Boolean;
  function eic(s1: AnsiString; s2: AnsiString): Boolean;
  function isDigit(s: AnsiString): Boolean;
  function isVar(s: AnsiString): Boolean;
  function isFunc(s: AnsiString): Boolean;
  function isOper(s: AnsiString): Boolean;
  function isNumOp(s: AnsiString): Boolean;
  function isCmpOp(s: AnsiString): Boolean;
  function isUOp(s: AnsiString): Boolean;
  function isSTok(s: AnsiString): Boolean;
  function isNumFunc(s: AnsiString): Boolean;
  function convert(vec: TStrings; loc: TStrings; loct: TStrings): AnsiString;
  function ripStr(s: AnsiString): AnsiString;
  function getTokens(text: AnsiString): TStrings;
  function getUpTo(vec: TStrings; be: Longint; d: AnsiString): TStrings;
  function skipTo(vec: TStrings; be: Longint; d: AnsiString) : Longint;
  function skipBack(vec: TStrings; be: Longint; d: AnsiString) : Longint;
  function skipExpr(vec: TStrings; be: Longint) : Longint;
  function skipPar(vec: TStrings; be: Longint) : Longint;
  function getRema(vec: TStrings; p: Longint): TStrings;
  function getChunk(vec: TStrings; b: Longint; e: Longint): TStrings;
  function getProcParms(text: AnsiString): TStrings;
  function getIndexOf(strs: TStrings; str: AnsiString): Longint;
  function getOpPre(str: AnsiString): Longint;
  function beginWith(strs: TStrings; str: AnsiString): Boolean;
  procedure throwError(l: Longint; t: Longint; msg: AnsiString; mes: TStrings);
//  procedure addToVar(name: AnsiString);
  procedure newVar(loc: TStrings; loct: TStrings; vName: AnsiString; vType: AnsiString);
  function convertArg(vec: TStrings; loc: TStrings; loct: TStrings; t: Longint): AnsiString;
  function convertProc(vec: TStrings; loc: TStrings; loct: TStrings): AnsiString;
  function convertFunc(vec: TStrings; loc: TStrings; loct: TStrings; t: Longint): AnsiString;
  function convertCmpOp(s: AnsiString): AnsiString;
  function toString(vec: TStrings): AnsiString;
  function toDeclaration(subLoc: TStrings; subTypes: TStrings; parms: TStrings): AnsiString;
  function convertExpr(vec: TStrings; loc: TStrings; loct: TStrings; t: Longint): AnsiString;
  function skipBlock(lines: TStrings; lNum: Longint; b: AnsiString; e: AnsiString): Longint;

var
  FormBuild: TFormBuild;
  prgName: AnsiString;
  varList: TStrings;
  typList: TStrings;
  errorAppend: Boolean;
  ops: TStrings;
  opers: TStrings;
  oPri: TStrings; 
  esp: TStrings;
  keyw: TStrings;
  opPar: TStrings;
  clPar: TStrings;
  consVec: TStrings;

  lNum: Longint;
  mes: TStrings;
  funcList: TStrings;
  prgList: TStrings;
  funcTxt: TStrings;
  prgTxt: TStrings;
  funcArgs: TStrings;
  prgArgs: TStrings;
  pList: TStrings;
  fList: TStrings;

  aCount: Longint;
  aNumR: TStrings;
  aNumC: TStrings;

  isPrg: Boolean;
  isSaveScreen: Boolean;
  strSize: Longint;
  isUseAll: Boolean;

  used: TStrings;
  newLine: AnsiString;


implementation

uses MainFrame;

{$R *.dfm}

function skipBlock(lines: TStrings; lNum: Longint; b: AnsiString; e: AnsiString): Longint;
var
  par: Longint;
  vec: TStrings;
begin
  par := 0;

  if (lNum < lines.Count) then vec := getTokens(lines[lNum]);

  while (lNum < lines.Count) and ((par > 0) or (vec.count = 0) or not eic(vec.strings[0], e)) do
  begin
    if (vec.Count > 0) and eic(vec.strings[0], b) then
    begin
      par := par + 1;
    end
    else if (vec.Count > 0) and eic(vec.strings[0], e) then
    begin
      par := par - 1;
    end;

    if (lNum < lines.Count) then vec := getTokens(lines[lNum]);
    lNum := lNum + 1;
  end;

  skipBlock := lNum;
end;

function convertCmpOp(s: AnsiString): AnsiString;
var
  str: AnsiString;
begin

  str := s;

  if (s = '=') then
  begin
    str := '==';
  end
  else if (s = '<>') then
  begin
    str := '!=';
  end;

  convertCmpOp := str;
end;

function isSTok(s: AnsiString): Boolean;
var
  vec : TStrings;
begin
  vec := getTokens(s);

  isSTok := (vec.Count = 1);
end;

function isUOp(s: AnsiString): Boolean;
begin
  isUOp := (s = '#') or (s = '@') or (s = 'not') or (s = '-neg-') or (s = '-pos-');
end;

function isCmpOp(s: AnsiString): Boolean;
begin
  isCmpOp := (s = '=') or (s = '>=') or (s = '>') or (s = '<=') or (s = '<') or (s = '!=');
end;

function isNumOp(s: AnsiString): Boolean;
begin
  isNumOp := (s = '+') or (s = '-') or (s = '*') or (s = '/');
end;

function getOpPre(str: AnsiString): Longint;
var
  i: Longint;
begin
  i := getIndexOf(opers, str);

  if (i >= 0) then
  begin
    i := strToint(oPri.Strings[i]);
  end;

  getOpPre := i;
end;

function isOper(s: AnsiString): Boolean;
begin
  isOper := (getIndexOf(opers, s) >= 0);
end;

function convertExpr(vec: TStrings; loc: TStrings; loct: TStrings; t: Longint): AnsiString;
var
  stk: TStrings;
  pil: TStrings;
  str: AnsiString;
  expr: AnsiString;
  pos: Longint;
  i: Longint;
  stmp: AnsiString;
  op: AnsiString;
  LOpe: AnsiString;
  ROpe: AnsiString;
  at: Longint;
begin

  at := 1;

  str := '';

  stk := TStringList.Create;
  pil := TStringList.Create;

  pil.Append('(');

  vec.Append(')');

  pos := 0;

  while (pil.Count > 0) do
  begin
    if (pos >= vec.Count) then
    begin
      throwError(lNum, pos, '-Expecting '')''.', mes);
      convertExpr := str; exit;
    end;

    //if (pos + 1 >= vec.Count) or (vec[pos+1] <> '(') then
    //begin
      i := skipExpr(vec, pos);
    (*end
    else
    begin
      i := pos + 1;
    end;  *)

    expr := toString(getChunk(vec, pos, i-1));

    if (expr = '(') then
    begin
      pil.Append('(');
    end
    else if (expr = ')') then
    begin
      stmp := pil.Strings[pil.Count - 1];
      pil.Delete(pil.Count - 1);
      while (stmp <> '(') do
      begin
        if (pil.Count <= 0) then
        begin
          throwError(lNum, pos, '-Expecting ''(''.', mes);
          convertExpr := str; exit;
        end;
        stk.Append(stmp);
        stmp := pil.Strings[pil.Count - 1];
        pil.Delete(pil.Count - 1);
      end;
    end
    else if isOper(expr) then
    begin
      if ((pos = 0) or (isOper(vec.Strings[pos - 1]) and (vec.Strings[pos - 1] <> ')') )) then
      begin
        if (vec.Strings[pos] = '-') then
        begin
          pil.Append('-neg-');
        end
        else if (vec.Strings[pos] = '+') then
        begin
          pil.Append('-pos-');
        end
        else
        begin
          pil.Append(expr);
        end;
      end
      else
      begin
        while (getOpPre(pil.Strings[pil.Count - 1]) >= getOpPre(expr)) do
        begin
          stk.Append(pil.Strings[pil.Count - 1]);
          pil.Delete(pil.Count - 1);
        end;
        pil.Append(expr);
      end;
    end
    else
    begin
      stk.Append(toString(getChunk(vec, pos, i - 1)));
    end;

    pos := i;
  end;

  pil.Clear;

  stk.Append(')');
  pos := 0;

  while (stk.Strings[pos] <> ')') do
  begin
    if isOper(stk.Strings[pos]) then
    begin

      //Ajouter vrifications

      op := stk.Strings[pos];

      if (pil.Count = 0) then
      begin
        throwError(lNum, pos, '-Invalid expression.', mes);
        convertExpr := str; exit;
      end;

  //    pos := pos + 1;
      Rope := pil.Strings[pil.Count - 1];
      pil.Delete(pil.Count - 1);

      if (pil.Count = 0) and not isUOp(op) then
      begin
        throwError(lNum, pos, '-Invalid expression.', mes);
        convertExpr := str; exit;
      end;

      if not isUOp(stk.Strings[pos]) then
      begin
   //     pos := pos + 1;
        Lope := pil.Strings[pil.Count - 1];
        pil.Delete(pil.Count - 1);
      end;

      if (op = '^') then
      begin
        at := 1;
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 1);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('pow ( ' + Lope + ', ' + Rope + ' ) ');
      end
      else if (isNumOp(op)) then
      begin
        at := 1;
        //pil.Append(' ( ' + convertArg(getTokens(Lope), loc, loct, 1) + op + convertArg(getTokens(Rope), loc, loct, 1) + ' ) ');
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 1);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('( ' + Lope + ' ' + op + ' ' + Rope + ' ) ');
      end
      else if (op = '&') then
      begin
        at := 0;
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 2);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 2);
        end;
        pil.Append('concatVariantStr ( ' + Lope + ', ' + Rope + ' ) ');
      end
      else if (op = '-neg-') then
      begin
        at := 1;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('- ' + Rope + ' ');
      end
      else if (op = '+pos+') then
      begin
        at := 1;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append(Rope + ' ');
      end
      else if (op = 'xor') then
      begin
        at := 1;
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 1);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('( ( ( BOOL ) ' + Lope + ' ) ^ ( ( BOOL ) ' + Rope + ' ) ) ');
      end
      else if (op = 'or') then
      begin
        at := 1;
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 1);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('( ' + Lope + ' || ' + Rope + ' ) ');
      end
      else if (op = 'and') then
      begin
        at := 1;
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 1);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('( ' + Lope + ' && ' + Rope + ' ) ');
      end
      else if isCmpOp(op) then
      begin
        at := 1;
        if isSTok(Lope) then
        begin
          Lope := convertArg(getTokens(Lope), loc, loct, 0);
        end;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 0);
        end;
        pil.Append('( variantCompareTo ( ' + Lope + ', ' + Rope + ' ) ' + convertCmpOp(op) + ' 0 ) ');
      end
      else if (op = 'not') then
      begin
        at := 1;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 1);
        end;
        pil.Append('! ' + Rope);
      end
      else if (op = '#') then
      begin
        at := 0;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 0);
        end;
        pil.Append('( * ' + Rope + '.ptr ) ');
      end
      else if (op = '@') then
      begin
        at := 3;
        if isSTok(Rope) then
        begin
          Rope := convertArg(getTokens(Rope), loc, loct, 0);
        end;
        pil.Append('( & ' + Rope + ' ) ');
      end;


    end
    else
    begin
      if isNumFunc(stk.Strings[pos]) or isDigit(stk.Strings[pos]) or eic(stk.Strings[pos], 'true') or eic(stk.Strings[pos], 'false') then
      begin
        at := 1;
        pil.Append(convertArg(getTokens(stk.Strings[pos]), loc, loct, 1));
      end
     (* else if (getIndexOf(consVec, stk.Strings[pos]) >= 0) or ... then
      begin
        at := 1;
        pil.Append(stk.Strings[pos]);
      end     *)
      else
      begin
        at := 0;
        pil.Append(convertArg(getTokens(stk.Strings[pos]), loc, loct, 0));
      end;
    end;

    pos := pos + 1;

  end;

  //Ajouter vrifications

    stmp := pil.Strings[pil.Count - 1];

    if (getIndexOf(consVec, stk.Strings[pos]) >= 0) or (getIndexOf(varList, stmp) >= 0) and (typList.Strings[getIndexOf(varList, stmp)] <> 'Variant') then
    begin
      stmp := stmp;
    end
    else if (at = 0) then
    begin
      if (t = 0) then
      begin
        stmp := stmp;
      end
      else if (t = 1) then
      begin
        stmp := stmp + '.num';
      end
      else if (t = 2) then
      begin
        stmp := stmp;
      end
      else if (t = 3) then
      begin
        stmp := ' & ' + stmp;
      end;
    end
    else if (at = 1) then
    begin
      if (t = 0) then
      begin
        stmp := ' newVariantNum ( ' + stmp + ' ) ';
      end
      else if (t = 1) then
      begin
        stmp := stmp;
      end
      else if (t = 2) then
      begin
        stmp := ' asVariantString ( newVariantNum ( ' + stmp + ' ) ) ';
      end
      else if (t = 3) then
      begin
        throwError(lNum, pos, '-Invalid argument.', mes);
        convertExpr := str; exit;
      end;
    end
    else if (at = 2) then
    begin
      if (t = 0) then
      begin
        stmp := stmp;
      end
      else if (t = 1) then
      begin
        stmp := 'asVariantNum ( ' + stmp + ' ) ';
      end
      else if (t = 2) then
      begin
        stmp := stmp;
      end
      else if (t = 3) then
      begin
        stmp := ' & ' + stmp;
      end;
    end
    else if (at = 3) then
    begin
      if (t = 0) then
      begin
        stmp := stmp;
      end
      else if (t = 1) then
      begin
        stmp := 'asVariantNum ( ' + stmp + ' ) ';
      end
      else if (t = 2) then
      begin
        stmp := 'asVariantString ( ' + stmp + ' ) ';
      end
      else if (t = 3) then
      begin
        stmp := ' & ' + stmp;
      end;
    end;

  convertExpr := stmp;
end;

function isNumFunc(s: AnsiString): Boolean;
var
  i: Longint;
  vec: TStrings;
begin
  isNumFunc := false;

  if (length(s) > 0) then
  begin
    s := getTokens(s)[0];
  end
  else
  begin
    exit;
  end;

  for i := 0 to fList.Count - 1 do
  begin
    vec := getTokens(fList.Strings[i]);
    
    if (0 < vec.Count) and eic(vec.Strings[1], s) and eic(vec.Strings[0], 'NUMERIC') then
    begin
      isNumFunc := true;
      break;
    end;
  end;
end;

function isFunc(s: AnsiString): Boolean;
var
  i: Longint;
  vec: TStrings;
begin
  isFunc := false;

  if (getIndexOf(funcList, s) >= 0) then
  begin
    isFunc := true;
    exit;
  end;

  for i := 0 to fList.Count - 1 do
  begin
    vec := getTokens(fList.Strings[i]);
    
    //if (0 < vec.Count) and eic(vec.Strings[0], s) then
    if (0 < vec.Count) and eic(vec.Strings[1], s) then
    begin
      isFunc := true;
      break;
    end;
  end;
end;

function skipPar(vec: TStrings; be: Longint) : Longint;
var
  par: Longint;
  pos: Longint;
begin
  par := 0;
  pos := be;

  while (be < vec.Count) and ((par > 0) or (pos = be)) do
  begin
    if (getIndexOf(opPar, vec.Strings[pos]) >= 0) then
    begin
      par := par + 1;
    end
    else if (getIndexOf(clPar, vec.Strings[pos]) >= 0) then
    begin
      par := par - 1;
    end;

    pos := pos + 1;
  end;

  skipPar := pos;
end;

function skipExpr(vec: TStrings; be: Longint) : Longint;
var
  par: Longint;
begin
  par := 0;

  if (be < vec.Count) and (getIndexOf(keyw, vec[be]) > 0) then
  begin
    be := be + 1;
    skipExpr := be;
    exit;
  end;

  be := be + 1;

  while (be < vec.Count) and ((par > 0) or (getIndexOf(opPar, vec.Strings[be]) >= 0)) do
  begin
    if (getIndexOf(opPar, vec.Strings[be]) >= 0) then
    begin
      par := par + 1;
    end
    else if (getIndexOf(clPar, vec.Strings[be]) >= 0) then
    begin
      par := par - 1;
    end;

    be := be + 1;
  end;

  skipExpr := be;
end;

function toDeclaration(subLoc: TStrings; subTypes: TStrings; parms: TStrings): AnsiString;
var
  i: Longint;
  str: AnsiString;
begin
  str := '';

  for i := 0 to subLoc.Count - 1 do
  begin
    if (getIndexOf(parms, subLoc.Strings[i]) < 0) then
    begin
      str := str + subTypes.Strings[i] + ' ' + UpperCase(subLoc.Strings[i]) + ' ; ';
    end;
  end;

  toDeclaration := str;
end;

function getChunk(vec: TStrings; b: Longint; e: Longint): TStrings;
var
  c: TStrings;
  i: Longint;
begin
  c := TStringList.Create;

  for i := b to e do
  begin
    if (i >= vec.Count) then
    begin
      break;
    end;
    c.Append(vec.Strings[i]);
  end;

  getChunk := c;
end;

function skipBack(vec: TStrings; be: Longint; d: AnsiString) : Longint;
var
  par: Longint;
  pos: Longint;
begin
  par := 0;
  pos := be;

  while (pos > 0) and ((par > 0) or (vec.Strings[pos] <> d)) do
  begin
    if (getIndexOf(clPar, vec.Strings[pos]) >= 0) then
    begin
      par := par + 1;
    end
    else if (getIndexOf(opPar, vec.Strings[pos]) >= 0) then
    begin
      par := par - 1;
    end;

    pos := pos - 1;
  end;

  skipBack := pos;
end;

function skipTo(vec: TStrings; be: Longint; d: AnsiString) : Longint;
var
  par: Longint;
  pos: Longint;
begin
  par := 0;
  pos := be;

  while (pos < vec.Count) and ((par > 0) or (vec.Strings[pos] <> d)) do
  begin
    if (getIndexOf(opPar, vec.Strings[pos]) >= 0) then
    begin
      par := par + 1;
    end
    else if (getIndexOf(clPar, vec.Strings[pos]) >= 0) then
    begin
      par := par - 1;
    end;

    pos := pos + 1;
  end;

  skipTo := pos;
end;

function toString(vec: TStrings): AnsiString;
var
  str: AnsiString;
  i: Longint;
begin
  str := '';

  for i := 0 to vec.count - 1 do
  begin
    if (i > 0) then
    begin
      str := str + ' ';
    end;
    str := str + vec.strings[i];
  end;

  toString := str;
end;

function convertProc(vec: TStrings; loc: TStrings; loct: TStrings): AnsiString;
var
  vecp: TStrings;
  par: TStrings;
  str : AnsiString;

  s: AnsiString;
  pos : Longint;
  i: Longint;
  fin : Longint;
  b: Boolean;
  last : AnsiString;
begin
  pos := 0;
  str := '';
  last := '';

  if (pos < vec.Count) then
  begin

    b := false;
    for i := 0 to pList.Count - 1 do
    begin
      vecp := getTokens(pList.Strings[i]);
      if eic(vecp.Strings[0], vec.Strings[pos]) then
      begin
        b := true;
        used.Append(vecp.Strings[0]);
        break;
      end;
    end;

    if b then
    begin
      i := 1;

      //pos := pos + 1;

      par := getProcParms(toString(getChunk(vec,1,vec.Count)));

      if (1 < vecp.Count) and (vecp.Strings[1] = '.') then
      begin
        last := vecp.Strings[2];
        for i := 0 to par.Count - 1 do
        begin
          if (last = 'POINTER') then
          begin
            if not isVar(par.Strings[i]) then
            begin
              throwError(lNum, pos, '- Expecting a variant as parameter of procedure ''' + vecp.Strings[0] + '''.', mes);
              convertProc := str; exit;
            end;
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 3) + ' ) ; ';
          end
          else if (last = 'NUMERIC') then
          begin
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 1) + ' ) ; ';
          end
          else if (last = 'STRING') then
          begin
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 2) + ' ) ; ';
          end
          else if (last = 'VARIANT') then
          begin
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 0) + ' ) ; ';
          end;
        end;

        convertProc := str; exit;
      end
      else
      begin
        str := str + UpperCase(vec.Strings[pos]) + ' (';
      end;

      if (par.Count < vecp.Count - 1) then
      begin
        for i := par.Count + 1 to vecp.Count - 1 do
        begin
          if (vecp.Strings[i] = 'VARIANT') or (vecp.Strings[i] = 'STRING') or (vecp.Strings[i] = 'NUMERIC') or (vecp.Strings[i] = 'POINTER') then
          begin
            break;
          end;
          par.Append(vecp.Strings[i]);
        end;
      end;

      if (vecp.Count - 1 <> par.Count) then
      begin
          throwError(lNum, pos, '-Invalid parameter count for procedure ''' + vecp.Strings[0] + '''.', mes);
          convertProc := str; exit;
      end;

      i := 1;

      pos := pos + 1;

      while (i < vecp.Count) do
      begin
        if (vecp.Strings[i] = 'POINTER') then
        begin
          if not isVar(par.Strings[i-1]) then
          begin
            throwError(lNum, pos, '- Expecting a variant as parameter of procedure ''' + vecp.Strings[0] + '''.', mes);
            convertProc := str; exit;
          end
          else if (getIndexOf(loc, vec.Strings[pos]) < 0) and (getIndexOf(varList, vec.Strings[pos]) < 0) then
          begin
            varList.Append(vec.Strings[pos]);
          end;

          if (i > 1) then
          begin
            str := str + ',';
          end;
          str := str + ' ' + convertExpr(getTokens(par.Strings[i-1]), loc, loct, 3);

          i := i + 1;
        end
        else
        begin

          if (i > 1) then
          begin
            str := str + ',';
          end;

          if (vecp.Strings[i] = 'STRING') then
          begin
            str := str + ' ' + convertExpr(getTokens(par.Strings[i-1]), loc, loct, 2);
          end
          else if (vecp.Strings[i] = 'VARIANT') then
          begin
            str := str + ' ' + convertExpr(getTokens(par.Strings[i-1]), loc, loct, 0);
          end
          else
          begin
            str := str + ' ' + convertExpr(getTokens(par.Strings[i-1]), loc, loct, 1);
          end;

          i := i + 1;
        end;
      end;

      str := str + ' ) ; ';

    end
    else if (pos + 1 < vec.Count) and (vec.Strings[pos + 1] = '(') and (skipBack(vec, vec.Count - 1, '->') = -1) and (getIndexOf(ops, vec.Strings[pos]) < 0) and not isFunc(vec.Strings[pos]) then
    begin
      str := str + UpperCase (vec.Strings[pos]) + ' ( ';

      pos := pos + 1;

      if (pos < vec.Count) then
      begin
        pos := pos + 1;
      end;

      par := getProcParms(toString(getChunk(vec,pos,vec.Count - 1)));

      for i := 0 to par.Count - 1 do
      begin
        if (i > 0) then
        begin
          str := str + ', ';
        end;

        str := str + convertExpr(getTokens(par.Strings[i]), loc, loct, 0);
      end;

      str := str + ' ; ' ;

      convertProc := str; exit;
    end;
  end;

  convertProc := str;
end;

function convertFunc(vec: TStrings; loc: TStrings; loct: TStrings; t: Longint): AnsiString;
var
  vecp: TStrings;
  par: TStrings;
  str : AnsiString;

  s: AnsiString;
  pos : Longint;
  i: Longint;
  fin : Longint;
  b: Boolean;
  last : AnsiString;
begin
  pos := 0;
  str := '';
  last := '';

  if (pos < vec.Count) then
  begin

    b := false;
    for i := 0 to fList.Count - 1 do
    begin
      vecp := getTokens(fList.Strings[i]);
      if eic(vecp.Strings[1], vec.Strings[pos]) then
      begin
        b := true;
        used.Append(vecp.Strings[1]);
        break;
      end;
    end;

    if b then
    begin
      i := 2;
      if (t = 0) then
      begin
        if (vecp.Strings[0] = 'NUMERIC') then
        begin
          str := str + 'newVariantNum ( ' + UpperCase(vec.Strings[pos]) + ' (';
        end
        else if (vecp.Strings[0] = 'STRING') then
        begin
          str := str + 'newVariantStr ( ' + UpperCase(vec.Strings[pos]) + ' (';
        end
        else
        begin

          str := str + UpperCase(vec.Strings[pos]) + ' (';

        end;
      end
      else if (t = 1) then
      begin
        if (vecp.Strings[0] = 'VARIANT') then
        begin
          str := str + UpperCase(vec.Strings[pos]) + ' (';
        end
        else
        begin

          str := str + UpperCase(vec.Strings[pos]) + ' (';

        end;
      end
      else
      begin

        str := str + UpperCase(vec.Strings[pos]) + ' (';

      end;

      pos := pos + 2;

      par := getProcParms(toString(getChunk(vec,pos,vec.Count-1)));

      if (2 < vecp.Count) and (vecp.Strings[2] = '.') then
      begin
        last := vecp.Strings[3];
        for i := 0 to par.Count - 1 do
        begin
          if (last = 'POINTER') then
          begin
            if not isVar(par.Strings[i]) then
            begin
              throwError(lNum, pos, '- Expecting a variant as parameter of function ''' + vecp.Strings[0] + '''.', mes);
              convertFunc := str; exit;
            end;
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 3) + ' ) ';
          end
          else if (last = 'NUMERIC') then
          begin
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 1) + ' ) ';
          end
          else if (last = 'STRING') then
          begin
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 2) + ' ) ';
          end
          else if (last = 'VARIANT') then
          begin
            str := str + UpperCase(vecp.Strings[0]) + ' ( ' + convertExpr(getTokens(par.Strings[i]), loc, loct, 0) + ' ) ';
          end;
        end;

        convertFunc := str; exit;
      end;

      if (par.Count < vecp.Count - 2) then
      begin
        for i := par.Count + 2 to vecp.Count - 1 do
        begin
          if (vecp.Strings[i] = 'VARIANT') or (vecp.Strings[i] = 'STRING') or (vecp.Strings[i] = 'NUMERIC') or (vecp.Strings[i] = 'POINTER') then
          begin
            break;
          end;
          par.Append(vecp.Strings[i]);
        end;
      end;

      if (vecp.Count - 2 <> par.Count) then
      begin
          throwError(lNum, pos, '-Invalid parameter count for function ''' + vecp.Strings[1] + '''.', mes);
          convertFunc := str; exit;
      end;

      i := 2;

      while (i < vecp.Count) do
      begin
        if (vecp.Strings[i] = 'POINTER') then
        begin
          if not isVar(par.Strings[i-2]) then
          begin
            throwError(lNum, pos, '- Expecting a variant as parameter of function ''' + vecp.Strings[0] + '''.', mes);
            convertFunc := str; exit;
          end
          else if (getIndexOf(loc, vec.Strings[pos]) < 0) and (getIndexOf(varList, vec.Strings[pos]) < 0) then
          begin
            varList.Append(vec.Strings[pos]);
          end;

          if (i > 2) then
          begin
            str := str + ',';
          end;
          str := str + UpperCase(' & ' + par.Strings[i-2]);

          i := i + 1;
        end
        else
        begin

          if (i > 2) then
          begin
            str := str + ',';
          end;

          if (vecp.Strings[i] = 'STRING') then
          begin
            str := str + ' ' + convertExpr(getTokens(par.Strings[i-2]), loc, loct, 2);
          end
          else if (vecp.Strings[i] = 'VARIANT') then
          begin
            str := str + ' ' + convertExpr(getTokens(par.Strings[i-2]), loc, loct, 0);
          end
          else
          begin
            str := str + ' ' + convertExpr(getTokens(par.Strings[i-2]), loc, loct, 1);
          end;

          i := i + 1;
        end;
      end;

      if (t = 0) then
      begin
        if (vecp.Strings[0] = 'NUMERIC') then
        begin
          str := str + ' ) ) ';
        end
        else if (vecp.Strings[0] = 'NUMERIC') then
        begin
          str := str + ' ) ) ';
        end
        else
        begin

          str := str + ' ) ';

        end;
      end
      else if (t = 1) then
      begin
        if (vecp.Strings[0] = 'VARIANT') then
        begin
          str := str + ' ).num ';
        end
        else
        begin

          str := str + ' )  ';

        end;
      end
      else
      begin

        str := str + ' )  ';

      end;



    end
    else if eic(vec.Strings[pos], 'when') then
    begin
      pos := pos + 2;
      par := getProcParms(toString(getChunk(vec,pos,vec.Count - 1)));
      if (par.Count <> 3) then
      begin
        throwError(lNum, pos, '-Invalid parameter count for ''when''.', mes);
        convertFunc := str; exit;
      end;

      str := str + '( (' + convertExpr(getTokens(par.Strings[0]), loc, loct, 1) + ') ? ' + convertExpr(getTokens(par.Strings[1]), loc, loct, t) + ' : ' + convertExpr(getTokens(par.Strings[2]), loc, loct, t) + ') ';

    end
    else if (pos + 1 < vec.Count) and (vec.Strings[pos + 1] = '(') then
    begin

      str := str + UpperCase (vec.Strings[pos]) + ' ( ';

      pos := pos + 1;

      if (pos < vec.Count) then
      begin
        pos := pos + 1;
      end;

      par := getProcParms(toString(getChunk(vec,pos,vec.Count - 1)));

      for i := 0 to par.Count - 1 do
      begin
        if (i > 0) then
        begin
          str := str + ', ';
        end;

        str := str + convertExpr(getTokens(par.Strings[i]), loc, loct, 0);
      end;

      str := str + ' ' ;

      if (t = 0) then
      begin
          str := str + ' ) ';
      end
      else if (t = 1) then
      begin

          str := str + ' ).num ';

      end;

      convertFunc := str; exit;
    end;
  end;

  convertFunc := str;
end;

function getRema(vec: TStrings; p: Longint): TStrings;
var
  tok : TStrings;
  i: Longint;
begin
  tok := TStringList.Create;

  for i := p to vec.Count - 1 do
  begin
    tok.Append(vec.Strings[i]);
  end;

  getRema := tok;
end;

function ripStr(s: AnsiString): AnsiString;
var
  str: AnsiString;
  i: Longint;
begin
  str := '';

  for i := 1 to length(s) do
  begin
    str := str + s[i];
  end;

  ripStr := str;
end;

function isVar(s: AnsiString): Boolean;
var
  i: Longint;
  n: Longint;
  rep: Boolean;
begin

  if (getIndexOf(ops, s) >= 0) then
  begin
    isVar := false;
    exit;
  end;

  rep := false;

  for i := 1 to 1 do
  begin
    rep := false;
    for n := 0 to 9 do
    begin
      if (s[i] = intToStr(n)) then
      begin
        rep := true;
        break;
      end;
    end;
    if rep then
    begin
      rep := false;
    end
    else
    begin
      rep := true;
    end;

    break;
  end;

  isVar := rep;
end;


function isDigit(s: AnsiString): Boolean;
var
  i: Longint;
  n: Longint;
  rep: Boolean;
begin

  rep := false;

  for i := 1 to length(s) do
  begin
    if (i = 1) and (s[1] = '-') then
    begin
      rep := true;
      continue;
    end;
    
    rep := false;
    for n := 0 to 9 do
    begin
      if (s[i] = intToStr(n)) then
      begin
        rep := true;
        break;
      end;
    end;
   // if not rep then
 //   begin
      break;
//    end;
  end;

  isDigit := rep;
end;

function convertArg(vec: TStrings; loc: TStrings; loct: TStrings; t: Longint): AnsiString;
var
  txtO : AnsiString;
  posL : Longint;
  wd: AnsiString;
  i: Longint;
  par: TStrings;
begin
  txtO := '';

  posL := 0;

  while (posL < vec.count) do
  begin
    wd := vec.Strings[posL];

    if (wd = '(') then
    begin
      i := skipPar(vec, posL);

      txtO := txtO + ' ( ' + convertExpr(getChunk(vec, posL + 1, i - 2), loc, loct, t) + ' ) ';

      posL := i - 1;
    end
    else if (posL + 1 < vec.Count) and (vec.Strings[posL + 1] = '[') then
    begin
        txtO := txtO + UpperCase(vec.Strings[posL]) + '.array->ptr';
        posL := posL + 1;
        i := skipPar(vec, posL);
        par := getProcParms(toString(getChunk(vec, posL + 1, i - 2)));

        if (par.Count < 1) then
        begin
          throwError(lNum, posL, '- Expecting index.', mes);
          convertArg := txtO; exit;
        end;

        if (par.Count < 2) then
        begin
          txtO := txtO + '[0]';
          txtO := txtO + '[ ( short int ) ' + convertExpr(getTokens(par[0]), loc, loct, 1) + ' - 1 ] ';
        end
        else
        begin
          txtO := txtO + '[ ( short int ) ' + convertExpr(getTokens(par[0]), loc, loct, 1) + ' - 1 ] ';
          posL := i;
          i := skipPar(vec, posL);
          txtO := txtO + '[ ( short int ) ' + convertExpr(getTokens(par[1]), loc, loct, 1) + ' - 1 ] ';
        end;


        posL := i - 1;
    end
    else if (posL + 1 < vec.Count) and (vec.Strings[posL + 1] = '(') and (getIndexOf(ops, wd) < 0) then
    begin
      i := skipPar(vec, posL + 1);

      txtO := txtO + convertFunc(getChunk(vec, posL, i - 2), loc, loct, t);

      posL := i - 1;
    end
    else if (1 <= length(wd)) then
    begin
      if (wd = '@') then
      begin
        posL := posL + 1;
        if (posL >= vec.Count) or not isVar(vec.Strings[posL]) then
        begin
          throwError(lNum, posL, '- Expecting variant for indirection.', mes);
          convertArg := txtO; exit;
        end;

        i := skipExpr(vec, posL);

        txtO := txtO + ' newVariantPtr ( & ' + convertExpr(getChunk(vec, posL, i - 1), loc, loct, 0) + ' ) ';
        posL := i;
      end
      else if (wd = '#') then
      begin
        posL := posL + 1;
        txtO := txtO + '* ';
        if (posL >= vec.Count) or not isVar(vec.Strings[posL]) then
        begin
          throwError(lNum, posL, '- Expecting variant for indirection.', mes);
          convertArg := txtO; exit;
        end;

        i := skipExpr(vec, posL);

        txtO := txtO + '(' + convertExpr(getChunk(vec, posL, i - 1), loc, loct, 0) + ').ptr ';
        posL := i;
      //  txtO := txtO + UpperCase(vec.Strings[posL]) + '.ptr';

     end
      else if (wd[1] = '"') then
      begin
        if (t = 0) then
        begin
          txtO := txtO + 'newVariantStr( ' + wd + ')';
        end
        else if (t = 1) then
        begin
          txtO := txtO + ripStr(wd);
        end
        else if (t = 2) then
        begin
//          txtO := txtO + wd;
          txtO := txtO + 'newVariantStr( ' + wd + ')';
        end
        else if (t = 3) then
        begin
          throwError(lNum, posL, '- Expecting variant for indirection.', mes);
          convertArg := txtO; exit;
        end;
      end
      else if (isDigit(wd) or eic(wd, 'true') or eic(wd, 'false')) then
      begin
        if eic(wd, 'true') then
        begin
          wd := '1';
        end
        else if eic(wd, 'false') then
        begin
          wd := '0';
        end;       

        if (t = 0) then
        begin
          txtO := txtO + 'newVariantNum( ' + wd + ')';
        end
        else if (t = 1) then
        begin
          txtO := txtO + wd;
        end
        else if (t = 2) then
        begin
        //  txtO := txtO + '"' + wd + '"';
          txtO := txtO + 'newVariantStr( ' + '"' + wd + '"' + ')';
        end
        else if (t = 3) then
        begin
          throwError(lNum, posL, '- Expecting variant for indirection.', mes);
          convertArg := txtO; exit;
        end;
      end
      else
      begin
      (*  if (getIndexOf(consVec, wd) >= 0) then
        begin
          txtO := txtO + wd;
          posL := posL + 1;
          continue;
        end;     *)

        if (getIndexOf(loc, wd) < 0) and (getIndexOf(varList, wd) < 0) and (getIndexOf(ops, wd) < 0) then
        begin
          varList.Append(wd);
          typList.Append('Variant');
        end;

        if (getIndexOf(varList, wd) >= 0) and (typList.Strings[getIndexOf(varList, wd)] <> 'Variant') then
        begin
          txtO := txtO + UpperCase(wd);
        end
        else if (t = 0) then
        begin
          txtO := txtO + UpperCase(wd);
        end
        else if (t = 1) then
        begin
          txtO := txtO + UpperCase(wd) + '.num';
        end
        else if (t = 2) then
        begin
          txtO := txtO + UpperCase(wd);
        end
        else if (t = 3) then
        begin
          txtO := txtO + '( & ' + UpperCase(wd) + ' ) ';
        end;
      end;
    end;

    posL := posL + 1;
  end;

  convertArg := txtO;
end;

function getUpTo(vec: TStrings; be: Longint; d: AnsiString): TStrings;
var
  tok: TStrings;
  par: Longint;
  pos: Longint;
begin
  tok := TStringList.Create;
  par := 0;
  pos := be;

  while (pos < vec.Count) and ((par > 0) or (vec.Strings[pos] <> d)) do
  begin
    if (getIndexOf(opPar, vec.Strings[pos]) >= 0) then
    begin
      par := par + 1;
    end
    else if (getIndexOf(clPar, vec.Strings[pos]) >= 0) then
    begin
      par := par - 1;
    end;

    tok.Append(vec.Strings[pos]);

    pos := pos + 1;
  end;

  getUpTo := tok;
end;

function getProcParms(text: AnsiString): TStrings;
var
  parms : TStrings;
  vec : TStrings;

  i: Longint;
  n : Longint;
begin
  parms := TStringList.Create;
  vec := getTokens(text);

  i := 0;

  while (i < vec.Count) and (vec.Strings[i] <> ')') do
  begin
    if (i > 0) then
    begin
      i := i + 1;
    end;

    if (i > vec.Count) then
    begin
      break;
    end;

    n := skipTo(vec, i, ',');
    parms.Append(toString(getChunk(vec, i, n-1)));
    i := n;
  end;

  getProcParms := parms;
end;

function eic(s1: AnsiString; s2: AnsiString): Boolean;
begin
  eic := (UpperCase(s1) = UpperCase(s2));
end;
procedure newVar(loc: TStrings; loct: TStrings; vName: AnsiString; vType: AnsiString);
var
  i: Longint;
begin

  for i := 0 to loc.Count - 1 do
  begin
    if (UpperCase(vName) = UpperCase(loc.Strings[i])) then
    begin
      loct.Strings[i] := vType;
      exit;
    end;
  end;

  for i := 0 to varList.Count - 1 do
  begin
    if (UpperCase(vName) = UpperCase(varList.Strings[i])) then
    begin
      typList.Strings[i] := vType;
      exit;
    end;
  end;

  varList.Append(vName);
  typList.Append(vType);

end;
(*
procedure addToVar(name: AnsiString);
var
  i: Longint;
begin

  for i := 0 to varList.Count - 1 do
  begin
    if (UpperCase(name) = UpperCase(varList.Strings[i])) then
    begin
      exit;
    end;
  end;

  varList.Append(name);
end;
*)
function convert(vec: TStrings; loc: TStrings; loct: TStrings): AnsiString;
var
  txtO: AnsiString;
  name: AnsiString;
  vName: AnsiString;
  vType: AnsiString;
  posL: Longint;

  strArgs: AnsiString;
  strSub: AnsiString;
  i: Longint;

  subLoc: TStrings;
  subTypes: TStrings;
  argLst: TStrings;

  subVec: TStrings;

  prg: Boolean;
  closed : Boolean;

  dlgW: AnsiString;
  dlgH: AnsiString;
  dlgX: AnsiString;
  dlgY: AnsiString;
  dlgRBuf: AnsiString;
  dlgPBuf: AnsiString;
  dlgTitle : AnsiString;
  dlgCompX: Longint;
  dlgCompY: Longint;
  dlgBody: AnsiString;
  dlgRCount : Longint;
  dlgRVar: TStrings;

  numR: Longint;
  numC: Longint;
  pVec: Longint;
begin

 // txtO := '';

  posL := 0;

  txtO := convertProc(vec, loc, loct);

  if (txtO <> '') then
  begin
    convert := txtO; exit;
  end;

  while (posL < vec.Count) do
  begin
    i := skipBack(vec, vec.Count - 1, '->');
    if (i >= 1) then
    begin
      if (i = vec.Count - 1) or not (isVar(vec[i+1]) or (vec[i+1] = '#')) then
      begin
        throwError(lNum, posL, '- Expecting variant after ''->''.', mes);
        convert := txtO; exit;
      end;

      if (vec.Strings[posL] = '[') then
      begin
        numR := 0;
        numC := 0;

        posL := posL + 1;

        strSub := convertExpr(getChunk(vec, i + 1, vec.Count - 1), loc, loct, 0);

        while ((vec.Strings[posL] <> ']') and (posL < i)) do
        begin
          if (vec.Strings[posL] = '[') then
          begin
            posL := posL + 1;
            while ((vec.Strings[posL] <> ']') and (posL < i)) do
            begin
              pVec := skipTo(vec, posL, ',');
              txtO := txtO + ' = '
            end;
          end;
        end;

        strSub := 'redimArray(' + strSub + '.array, ' + intToStr(numR) + ', ' + intToStr(numC) + ') ;';

        txtO := strSub + txtO;

        convert := txtO; exit;
      end
      else
      begin
        strSub := toString(getChunk(vec, i + 1, vec.Count - 1));

        txtO := txtO + convertExpr(getChunk(vec, i + 1, vec.Count - 1), loc, loct, 0) + ' = ' + convertExpr(getChunk(vec, 0, i - 1), loc, loct, 0) + ' ; ' ;
        posL := vec.Count;
      end;
    end
    else if eic(vec.Strings[posL], '''') then
    begin
      convert := txtO; exit;
    end
    else if eic(vec.Strings[posL], 'Option') then
    begin
      posL := posL + 1;
      if (posL >= vec.Count) then
      begin
          throwError(lNum, posL, '- Expecting option after ''Option''.', mes);
          convert := txtO; exit;
      end;


      if eic(vec.Strings[posL], 'NoSaveScreen') then
      begin
        isSaveScreen := false;
      end
      else if eic(vec.Strings[posL], 'UseAll') then
      begin
        isUseAll := true;
      end
      else if eic(vec.Strings[posL], 'StringSize') then
      begin
        posL := posL + 1;
        if (posL >= vec.Count) then
        begin
          throwError(lNum, posL, '- Expecting size after ''StringSize''.', mes);
          convert := txtO; exit;
        end;
        strSize := strToint(inttostr(strtoint(vec.Strings[posL]) + 1));
      end
      else
      begin
        throwError(lNum, posL, '- Token ''' + vec.Strings[posL] + ''' is not a valid option.', mes);
        convert := txtO; exit;
      end;

    end
    else if (eic(vec.Strings[posL], 'EndIf') or eic(vec.Strings[posL], 'EndWhile') or eic(vec.Strings[posL], 'EndFor') or eic(vec.Strings[posL], 'EndLoop')) then
    begin

      if (posL + 1 < vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected token after ''' + vec.Strings[posL] + '''.', mes);
        convert := txtO; exit;
      end;
        txtO := txtO + ' } ';
    end
    else if (eic(vec.Strings[posL], 'Call')) then
    begin

      if (posL + 1 >= vec.Count) then
      begin
        throwError(lNum, posL, '-Expecting instruction after ''Call''.', mes);
        convert := txtO; exit;
      end;
        txtO := toString(getChunk(vec, 1, vec.Count - 1)) + ' ; ';
        posL := vec.Count;
    end
    else if (eic(vec.Strings[posL], 'Else')) then
    begin

      if (posL + 1 < vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected token after ''Else''.', mes);
        convert := txtO; exit;
      end;
        txtO := txtO + ' } else { ';
    end
    else if (eic(vec.Strings[posL], 'Try')) then
    begin

      if (posL + 1 < vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected token after ''Try''.', mes);
        convert := txtO; exit;
      end;
      txtO := txtO + 'TRY ';
      i := 0;

      lNum := lNum + 1;

      if (lNum >= Form1.txtCode.Lines.Count) then
      begin
        throwError(lNum, posL, '-Expecting instructions after ''Try''.', mes);
        convert := txtO; exit;
      end;

      vec := getTokens(Form1.txtCode.Lines[lNum]);

      while (lNum < Form1.txtCode.Lines.Count) and ((i > 0) or (vec.Count < 1) or (not eic(vec[0], 'Else') and not eic(vec[0], 'EndTry'))) do
      begin

        if (vec.Count > 0) and eic(vec[0], 'If') then
        begin
          while (lNum < Form1.txtCode.Lines.Count) and ((i > 0) or (vec.Count < 1) or (not eic(vec[0], 'EndIf'))) do
          begin
            if (vec.Count > 0) and eic(vec[0], 'If') then
            begin
              i := i + 1;
            end
            else if (vec.Count > 0) and eic(vec[0], 'EndIf') then
            begin
              i := i - 1;
              if (i = 0) then
              begin
                break;
              end;
            end;

            txtO := txtO + convert(vec, loc, loct);

            lNum := lNum + 1;
            if (lNum >= Form1.txtCode.Lines.Count) then
            begin
              throwError(lNum, posL, '-Expecting instructions after ''Try''.', mes);
              convert := txtO; exit;
            end;
             vec := getTokens(Form1.txtCode.Lines[lNum]);
          end;
          continue;
        end;

        txtO := txtO + convert(vec, loc, loct);

        lNum := lNum + 1;
        if (lNum >= Form1.txtCode.Lines.Count) then
        begin
          throwError(lNum, posL, '-Expecting instructions after ''Try''.', mes);
          convert := txtO; exit;
        end;
        vec := getTokens(Form1.txtCode.Lines[lNum]);
      end;

      if (vec.Count > 0) and eic(vec[0], 'Else') then
      begin
        txtO := txtO + 'ONERR ';
      end
      else if (vec.Count > 0) and eic(vec[0], 'Endtry') then
      begin
        txtO := txtO + 'ONERR ENDTRY ';
      end;

      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Endtry')) then
    begin

      if (posL + 1 < vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected token after ''Endtry''.', mes);
        convert := txtO; exit;
      end;
        txtO := txtO + 'ENDTRY ';
    end
    else if (eic(vec.Strings[posL], 'PassErr')) then
    begin

      if (posL + 1 < vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected token after ''PassErr''.', mes);
        convert := txtO; exit;
      end;
        txtO := txtO + 'PASS ';
    end
    else if (eic(vec.Strings[posL], 'Exit')) then
    begin
        txtO := txtO + 'break ; ';
        convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Dialog')) then
    begin
        lNum := lNum + 1;

        dlgRBuf := '0';      
        dlgPBuf := '0';
        dlgY := 'CENTER';
        dlgX := 'CENTER';
        dlgW := 'LCD_WIDTH - 8';
        dlgH := '0';

        dlgTitle := 'newVariantStr("")';
        dlgBody := '';

        dlgCompY := 0;
        dlgCompX := 2;
        dlgRCount := 0;
        dlgRVar := TStringList.Create;

        while (lNum < Form1.txtCode.Lines.Count) do
        begin
          vec := getTokens(Form1.txtCode.Lines[lNum]);

          if (vec.Count > 0) then
          begin
            if eic(vec[0], 'Text') then
            begin
              dlgBody := dlgBody + 'Variant reserved_Dialog_Text' + intToStr(dlgCompY) + ' = ' + convertExpr(getChunk(vec, 1, vec.Count - 1), loc, loct, 2) + '; ';
              dlgBody := dlgBody + 'DialogAddText (reserved_Dialog_Handle, ' + intToStr(dlgCompX) + ', ' + intToStr((dlgCompY + 1) * 10 + 4) + ', reserved_Dialog_Text' + intToStr(dlgCompY) + '.str) ; ';
              dlgCompY := dlgCompY + 1;
            end
            else if eic(vec[0], 'Request') then
            begin
              dlgRBuf := inttostr(strtoint(dlgRBuf) + 1);
              subVec := getProcParms(toString(getChunk(vec, 1, vec.Count - 1)));

              if (subVec.Count <> 2) then
              begin
                throwError(lNum, posL, '-Invalid parameter count for ''Request''.', mes);
                convert := txtO; exit;
              end;

              dlgBody := dlgBody + 'Variant reserved_Dialog_Text' + intToStr(dlgCompY) + ' = ' + convertExpr(getTokens(subVec[0]), loc, loct, 2) + '; ';
              dlgBody := dlgBody + 'strcpy(&reserved_Dialog_RequestBuffer[' + inttostr(dlgRCount) + ' * VARIANT_STRING_SIZE], ' + UpperCase(subVec[1]) + '.str) ; ';
              dlgBody := dlgBody + 'strcat(reserved_Dialog_Text' + intToStr(dlgCompY) + '.str, ":") ; ';
              dlgBody := dlgBody + convertExpr(getTokens(subVec[1]), loc, loct, 0) + ' = asVariantString(' + convertExpr(getTokens(subVec[1]), loc, loct, 0) + ');';
              dlgBody := dlgBody + 'DialogAddRequest (reserved_Dialog_Handle, ' + intToStr(dlgCompX) + ', ' + intToStr((dlgCompY + 1) * 10 + 4) + ', reserved_Dialog_Text' + intToStr(dlgCompY) + '.str';
              dlgRVar.append(subVec[1]);
              dlgBody := dlgBody + ', ' + inttostr(dlgRCount) + ' * VARIANT_STRING_SIZE, VARIANT_STRING_SIZE - 1, ' + '(LCD_WIDTH - 18) / 6 - strlen(reserved_Dialog_Text' + intToStr(dlgCompY) + '.str)) ; ';
              dlgCompY := dlgCompY + 1;
              dlgRCount := dlgRCount + 1;
            end
            else if eic(vec[0], 'Title') then
            begin
              dlgTitle := convertExpr(getChunk(vec, 1, vec.Count - 1), loc, loct, 2);
            end
            else if eic(vec[0], 'EndDlog') then
            begin
              break;
            end
            else if eic(vec[0], '''') then
            begin
              (* nop *);
            end
            else
            begin
              throwError(lNum, posL, '-Expecting ''Title'', ''Text'', ''Request'', ''DropDown'' or ''EndDlog'' but ''' + vec[0] + ''' found.', mes);
              convert := txtO; exit;
            end;
          end;
          lNum := lNum + 1;
        end;

        dlgH := inttostr((dlgCompY + 3) * 10);

        txtO := txtO + '{ ';

        txtO := txtO + 'HANDLE reserved_Dialog_Handle = DialogNewSimple ( ' + dlgW + ', ' + dlgH + ' ); ';
        txtO := txtO + 'char *reserved_Dialog_RequestBuffer = malloc ( ' + dlgRBuf + ' * VARIANT_STRING_SIZE * sizeof(char) ) ; ';
        txtO := txtO + 'short *reserved_Dialog_PulldownBuffer = malloc ( ' + dlgPBuf + ' * sizeof(short)) ; ';
        txtO := txtO + 'Variant reserved_Dialog_Title = ' + dlgTitle + ' ; ';
        txtO := txtO + 'reserved_Dialog_RequestBuffer[0] = ''\0'';' ;

        txtO := txtO + 'DialogAddTitle (reserved_Dialog_Handle, reserved_Dialog_Title.str, BT_YES, BT_NO);' + newLine;
        txtO := txtO + dlgBody + ' ';

        txtO := txtO + 'OK = newVariantNum ( ( DialogDo (reserved_Dialog_Handle, ' + dlgX + ', ' + dlgY + ', reserved_Dialog_RequestBuffer, reserved_Dialog_PulldownBuffer) == KEY_ENTER ) ) ; ' + newLine;

        for i := 0 to dlgRVar.Count - 1 do
        begin
           txtO := txtO + UpperCase(dlgRVar[i]) + ' = newVariantStr ( &reserved_Dialog_RequestBuffer[' + inttoStr(i) + ' * VARIANT_STRING_SIZE] ) ; ';
        end;

        txtO := txtO + 'HeapFree ( reserved_Dialog_Handle ) ; ';
        txtO := txtO + 'free ( reserved_Dialog_RequestBuffer ) ; ';
        txtO := txtO + 'free ( reserved_Dialog_PulldownBuffer ) ; ';

        txtO := txtO + '} ';

        if not (eic(vec.Strings[posL], 'EndDlog')) then
        begin
          throwError(lNum, posL, '-Expecting EndDlog.', mes);
          convert := txtO; exit;
        end;
    end
    else if (eic(vec.Strings[posL], 'InputStr')) then
    begin
      posL := posL + 1;
      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected variant after ''InputStr''.', mes);
        convert := txtO; exit;
      end;

      if (posL + 2 >= vec.Count) then
      begin
        txtO := 'printf("?\n") ; getsn ((' + convertExpr(getChunk(vec, posL, vec.Count-1), loc, loct, 0) + ').str, VARIANT_STRING_SIZE) ; printf("\n");'
      end
      else
      begin
        i := skipTo(vec, posL, ',');
        if (i + 1 >= vec.Count) then
        begin

        end;
        txtO := 'printVariant(' + convertExpr(getChunk(vec, posL, i-1), loc, loct, 0) + ') ; printf("\n"); getsn ((' + convertExpr(getChunk(vec, i+1, vec.Count-1), loc, loct, 0) + ').str, VARIANT_STRING_SIZE) ;  printf("\n");';
        convert := txtO; exit;
      end;
      //txtO := txtO + 'continue ; '; //WTF ???
    end
    else if (eic(vec.Strings[posL], 'Input')) then
    begin
      posL := posL + 1;
      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected variant after ''Input''.', mes);
        convert := txtO; exit;
      end;

      if (posL + 2 >= vec.Count) then
      begin
        txtO := 'printf("?\n") ; getInput(& ' + convertExpr(getChunk(vec, posL, vec.Count-1), loc, loct, 0) + ') ; printf("\n");'
      end
      else
      begin
        i := skipTo(vec, posL, ',');
        if (i + 1 >= vec.Count) then
        begin

        end;
        txtO := 'printVariant(' + convertExpr(getChunk(vec, posL, i-1), loc, loct, 0) + ') ; printf("\n"); getInput(& ' + convertExpr(getChunk(vec, i+1, vec.Count-1), loc, loct, 0) + ') ; printf("\n");';
        convert := txtO; exit;
      end;
      //txtO := txtO + 'continue ; '; //WTF ???
    end
    else if (eic(vec.Strings[posL], 'Cycle')) then
    begin
        txtO := txtO + 'continue ; ';
        convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Loop')) then
    begin

      if (posL + 1 < vec.Count) then
      begin
        throwError(lNum, posL, '-Unexpected token after ''Loop''.', mes);
        convert := txtO; exit;
      end;
        txtO := txtO + ' while ( TRUE ) { ';
    end
    else if (eic(vec.Strings[posL], 'ElseIf')) then
    begin
      i := skipTo(vec, posL, 'then');

      if (posL + 1 >= vec.Count) then
      begin
        throwError(lNum, posL, '-Expecting condition after ''ElseIf''.', mes);
        convert := txtO; exit;
      end;
      if (i < vec.Count) then
      begin
        txtO := txtO + ' } else if ( ' + convertExpr(getChunk(vec, posL + 1, i - 1), loc, loct, 1) + ' ) { ';
      end
      else
      begin

        if (lNum + 1 >= Form1.txtCode.Lines.Count) then
        begin
          throwError(lNum, posL, '-Expecting instruction after ''ElseIf''.', mes);
          convert := txtO; exit;
        end;

        txtO := txtO + 'else if ( ' + convertExpr(getChunk(vec, posL + 1, i - 1), loc, loct, 1) + ' ) ';
      end;
        posL := 0;
        convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'While')) then
    begin

      if (posL + 1 >= vec.Count) then
      begin
        throwError(lNum, posL, '-Expecting condition after ''ElseIf''.', mes);
        convert := txtO; exit;
      end;
      txtO := txtO + 'while ( ' + convertExpr(getChunk(vec, posL + 1, vec.Count - 1), loc, loct, 1) + ' ) { ';
      posL := 0;
      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Pause')) then
    begin

      if (posL + 1 >= vec.Count) then
      begin
        txtO := txtO + 'ngetchx ( ) ; ';
      end
      else
      begin
        txtO := txtO + 'printVariant ( ' + convertExpr(getChunk(vec, posL + 1, vec.Count - 1), loc, loct, 0) + ') printf("\n") ; ngetchx ( ) ; ';
      end;
      posL := 0;
      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'For')) then
    begin
      subVec := getProcParms(toString(getChunk(vec, 1, vec.Count - 1)));

      if (subVec.Count > 4) or (subVec.Count < 3) then
      begin
        throwError(lNum, posL, '-Invalid parameter count for ''For''.', mes);
        convert := txtO; exit;
      end;

      txtO := txtO + 'for ( ' + convertArg(getTokens(subVec[0]), loc, loct, 1) + ' = ' + convertExpr(getTokens(subVec[1]), loc, loct, 1) + ' ; ' + convertArg(getTokens(subVec[0]), loc, loct, 1) + ' <= ' + convertExpr(getTokens(subVec[2]), loc, loct, 1) + ' ; ' + convertArg(getTokens(subVec[0]), loc, loct, 1);

      if (subVec.Count = 4) then
      begin
        txtO := txtO + '+= ' + convertExpr(getTokens(subVec[3]), loc, loct, 1) + ' ) { ';
      end
      else
      begin
        txtO := txtO + '++ ) { ';
      end;

      posL := 0;
      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'If')) then
    begin
      i := skipTo(vec, posL, 'then');

      if (posL + 1 >= vec.Count) then
      begin
        throwError(lNum, posL, '-Expecting condition after ''If''.', mes);
        convert := txtO; exit;
      end;
      if (i < vec.Count) then
      begin
        txtO := txtO + 'if ( ' + convertExpr(getChunk(vec, posL + 1, i - 1), loc, loct, 1) + ' ) { ';
      end
      else
      begin
        if (lNum + 1 >= Form1.txtCode.Lines.Count) then
        begin
          throwError(lNum, posL, '-Expecting instruction after ''If''.', mes);
          convert := txtO; exit;
        end;

        txtO := txtO + 'if ( ' + convertExpr(getChunk(vec, posL + 1, i - 1), loc, loct, 1) + ' ) ';
      end;
      
      posL := 0;
      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Stop')) then
    begin
      txtO := 'command_Stop ( ) ; ';
    end
    else if (eic(vec.Strings[posL], 'Return')) then
    begin
      txtO := txtO + 'return ';
      posL := posL + 1;
      txtO := txtO + convertExpr(getChunk(vec, posL, vec.count - 1), loc, loct, 0) + ' ; ';
      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Local')) then
    begin
      posL := posL + 1;
      //txtO := txtO + 'Variant ';
      loc.Append(vec.Strings[posL]);
      loct.Append('Variant');
      //txtO := txtO + UpperCase(vec.Strings[posL]);
      posL := posL + 1;
      while (posL < vec.Count) do
      begin
        if (vec.Strings[posL] <> ',') then
        begin
          throwError(lNum, posL, '- Expecting '',''.', mes);
          convert := txtO; exit;
        end;
        posL := posL + 1;
        if (posL >= vec.Count) then
        begin
          throwError(lNum, posL, '- Expecting variant after '',''.', mes);
          convert := txtO; exit;
        end;
        loc.Append(vec.Strings[posL]);
        loct.Append(vec.Strings[posL]);
        //txtO := txtO + ', ' + UpperCase(vec.Strings[posL]);
        posL := posL + 1;
      end;
      //txtO := txtO + ' ; ';
    end
    else if (eic(vec.Strings[posL], 'Export')) then
    begin
      posL := posL + 1;
      txtO := txtO + 'exportVariant ( & ' + UpperCase(vec.Strings[posL]) + ' , (char *) "' + UpperCase(vec.Strings[posL]) + '" ) ; ';
      posL := posL + 1;
      while (posL < vec.Count) do
      begin
        if (vec.Strings[posL] <> ',') then
        begin
          throwError(lNum, posL, '- Expecting '',''.', mes);
          convert := txtO; exit;
        end;
        posL := posL + 1;
        if (posL >= vec.Count) then
        begin
          throwError(lNum, posL, '- Expecting variant after '',''.', mes);
          convert := txtO; exit;
        end;
        txtO := txtO + 'exportVariant ( & ' + UpperCase(vec.Strings[posL]) + ' , (char *) "' + UpperCase(vec.Strings[posL]) + '" ) ; ';
        posL := posL + 1;
      end;
    end
    (*else if (eic(vec.Strings[posL], 'Exec')) then
    begin
      posL := posL + 1;
      if (posL < vec.Count) then
      begin
        vec := getRema(vec, posL);
        txtO := txtO + 'command_Exec (' + convertExpr(vec, loc, loct, 2) + ' ) ; ';
      end
      else
      begin
        throwError(lNum, posL, '- Expecting argument after ''Exec''.', mes);
      end;
      posL := vec.Count;
      convert := txtO; exit;
    end       *)
    else if (eic(vec.Strings[posL], 'New')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting variable name after ''New''.', mes);
        convert := txtO; exit;
      end;

      vName := vec.Strings[posL];

      posL := posL + 1;

      if (posL < vec.Count) then
      begin
        if (posL + 1 >= vec.Count) or not eic(vec.Strings[posL], 'As') then
        begin
          throwError(lNum, posL, '- Expecting ''As [type]'' after variable declaration.', mes);
          convert := txtO; exit;
        end;
        vType := toString(getRema(vec, posL + 1));

        posL := vec.Count;
      end
      else
      begin
        vType := 'Variant';
      end;

      newVar(loc, loct, vName, vType);
      //loc.Append(vName);
      //loct.Append(vType);
    end
    else if (eic(vec.Strings[posL], 'NewList')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting variable name after ''NewList''.', mes);
        convert := txtO; exit;
      end;

      vName := vec.Strings[posL];

      posL := posL + 1;

      if (posL + 1 >= vec.Count) or (vec.Strings[posL] <> ',') then
      begin
        throwError(lNum, posL, '- Expecting '', <nb elem>''.', mes);
        convert := txtO; exit;
      end;

      posL := posL + 1;



      newVar(loc, loct, vName, 'Variant');
      //loc.Append(vName);
      //loct.Append('Variant');
      txtO := txtO + UpperCase(vName) + ' = newVariantArray ( & array_' + intToStr(aCount) + ' ) ; ';

      aNumR.Append('1');
      aNumC.Append(vec.Strings[posL]);

 //     posL := posL + 1;
      aCount := aCount + 1;
    end
    else if (eic(vec.Strings[posL], 'NewMat')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting variable name after ''NewMat''.', mes);
        convert := txtO; exit;
      end;

      vName := vec.Strings[posL];

      posL := posL + 1;

      if (posL + 3 >= vec.Count) or (vec.Strings[posL] <> ',') or (vec.Strings[posL+2] <> ',') then
      begin
        throwError(lNum, posL, '- Expecting '', <nb row>[, <nb col>]''.', mes);
        convert := txtO; exit;
      end;

      posL := posL + 1;



      newVar(loc, loct, vName, 'Variant');
      //loc.Append(vName);
      //loct.Append('Variant');
      txtO := txtO + UpperCase(vName) + ' = newVariantArray ( & array_' + intToStr(aCount) + ' ) ; ';

      aNumR.Append(vec.Strings[posL]);
      posL := posL + 2;
      aNumC.Append(vec.Strings[posL]);

 //     posL := posL + 1;
      aCount := aCount + 1;
    end
    (*else if (eic(vec.Strings[posL], 'ReDim')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting variable name after ''ReDim''.', mes);
        convert := txtO; exit;
      end;

      vName := vec.Strings[posL];

      posL := posL + 1;

      if (posL + 1 >= vec.Count) or (vec.Strings[posL] <> ',') then
      begin
        throwError(lNum, posL, '- Expecting '', [size]''.', mes);
        convert := txtO; exit;
      end;
      posL := posL + 1;

    end *)
    else if (eic(vec.Strings[posL], 'Import')) then
    begin
      posL := posL + 1;
      newVar(loc, loct, vec.Strings[posL], 'Variant');
      txtO := txtO + 'importVariant ( & ' + UpperCase(vec.Strings[posL]) + ' , (char *) "' + UpperCase(vec.Strings[posL]) + '" ) ; ';
      posL := posL + 1;
      while (posL < vec.Count) do
      begin
        if (vec.Strings[posL] <> ',') then
        begin
          throwError(lNum, posL, '- Expecting '',''.', mes);
        end;
        posL := posL + 1;
        if (posL >= vec.Count) then
        begin
          throwError(lNum, posL, '- Expecting variant after '',''.', mes);
          convert := txtO; exit;
        end;
        newVar(loc, loct, vec.Strings[posL], 'Variant');
        txtO := txtO + 'importVariant ( & ' + UpperCase(vec.Strings[posL]) + ' , (char *) "' + UpperCase(vec.Strings[posL]) + '" ) ; ';
        posL := posL + 1;
      end;
    end
    else if (eic(vec.Strings[posL], 'Define')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting prgram or function name after ''Define''.', mes);
        convert := txtO; exit;
      end;

      name := vec.Strings[posL];

      posL := posL + 1;

        if (vec.Strings[posL] <> '(') then
        begin
          throwError(lNum, posL, '- Expecting ''(''.', mes);
          convert := txtO; exit;
        end;

        posL := posL + 1;

        subLoc := TStringList.Create;
        argLst := TStringList.Create;
        subTypes := TStringList.Create;
        strArgs := '';

        if (vec.Strings[posL] <> ')') then
        begin

          while (vec.Strings[posL] <> ')') do
          begin
            subLoc.Append(vec.Strings[posL]);
            SubTypes.Append('Variant');
            strArgs := strArgs + vec.Strings[posL] + ' ';
            argLst.Append(vec.Strings[posL]);

            posL := posL + 1;

            if (vec.Count <= posL) then
            begin
              throwError(lNum, posL, '- Declaration invalid.', mes);
              convert := txtO; exit;
            end;

            if (vec.Strings[posL] <> ',') and (vec.Strings[posL] <> ')') then
            begin
              throwError(lNum, posL,  '- Expecting '','' or '')''.', mes);
              convert := txtO; exit;
            end;

            if (vec.Strings[posL] = ')') then
            begin
              break;
            end;

            posL := posL + 1;

            if (vec.Count <= posL) then
            begin
              throwError(lNum, posL,  '- Declaration invalid.', mes);
              convert := txtO; exit;
            end;
          end;
        end;

        if (vec.Strings[posL] <> ')') then
        begin
          throwError(lNum, posL, '- Expecting '')''.', mes);
          convert := txtO; exit;
        end
        else
        begin
          posL := posL + 1;
        end;

      if (posL >= vec.Count) or not eic(vec.Strings[posL], '=') then
      begin
        throwError(lNum, posL, '- Expecting ''=''.', mes);
        convert := txtO; exit;
      end;

      posL := posL + 1;

      if (posL >= vec.Count) or (eic(vec.Strings[posL], 'Prgm') and eic(vec.Strings[posL], 'Func')) then
      begin
        throwError(lNum, posL, '- Expecting ''Prgm'' or ''Func''.', mes);
        convert := txtO; exit;
      end;

      if eic(vec.Strings[posL], 'Prgm') then
      begin
        prg := true;
        prgArgs.Append(strArgs);
        prgList.Append(name);
      end
      else
      begin
        prg := false;
        funcArgs.Append(strArgs);
        funcList.Append(name);
      end;

      lNum := lNum + 1;
      strSub := '';
      posL := 0;

      while (lNum < Form1.txtCode.Lines.Count) do
      begin
        vec := getTokens(Form1.txtCode.Lines[lNum]);

        if (vec.count > 0) and (eic(vec.strings[0], 'EndPrgm') or eic(vec.strings[0], 'EndFunc')) then
        begin
          break;
        end;

        strSub := strSub + convert(vec, subLoc, subTypes) + newLine;

        lNum := lNum + 1;
        posL := 0;
      end;

      if (vec.count <= 0) or not (eic(vec.strings[0], 'EndPrgm') or eic(vec.strings[0], 'EndFunc')) then
      begin
        throwError(lNum, posL, '- Expecting ''EndPrgm'' or ''EndFunc''.', mes);
        convert := txtO; exit;
      end;

      strSub := toDeclaration(subLoc, subTypes, argLst) + strSub;

      if prg then
      begin
        prgTxt.Append(strSub); 
      end
      else
      begin
        funcTxt.Append(strSub);
      end;

      convert := txtO; exit;

    end
    else if eic(vec.Strings[posL], 'Disp') then
    begin
      posL := posL + 1;

      if (posL >= vec.count) then
      begin
        throwError(lNum, posL, '- Expecting parameter after ''Disp''.', mes);
        convert := txtO; exit;
      end;

      strSub := '';

      for i := posL to vec.count - 1 do
      begin
        strSub := strSub + vec.Strings[i];
      end;

      vec := getProcParms(strSub);
      for i := 0 to vec.count - 1 do
      begin
        txtO := txtO + 'printVariant( ' + convertExpr(getTokens(vec.Strings[i]), loc, loct, 0) + ' ) ; ';
      end;

      txtO := txtO + 'printf ( "\n" ) ; ';

      convert := txtO; exit;
    end
    else if eic(vec.Strings[posL], 'Prompt') then
    begin
      posL := posL + 1;

      if (posL >= vec.count) then
      begin
        throwError(lNum, posL, '- Expecting parameter after ''Prompt''.', mes);
        convert := txtO; exit;
      end;

      strSub := '';

      for i := posL to vec.count - 1 do
      begin
        strSub := strSub + vec.Strings[i];
      end;

      vec := getProcParms(strSub);
      for i := 0 to vec.count - 1 do
      begin
        txtO := txtO + 'printf("' + toString(getTokens(vec.Strings[i])) + '?\n") ; ; getInput( &' + convertExpr(getTokens(vec.Strings[i]), loc, loct, 0) + ' ) ; printf("\n"); ';
      end;

      //txtO := txtO + 'printf ( "\n" ) ; ';

      convert := txtO; exit;
    end
    else if eic(vec.Strings[posL], 'Native') then
    begin
      lNum := lNum + 1;
      posL := 0;
      
      while (lNum < Form1.txtCode.Lines.Count) do
      begin
        vec := getTokens(Form1.txtCode.Lines[lNum]);

        if (vec.count > 0) and eic(vec.strings[0], 'EndNative') then
        begin
          break;
        end;

        txtO := txtO + Form1.txtCode.Lines[lNum];

        lNum := lNum + 1;
        posL := 0;
      end;

      if (vec.count <= 0) or not eic(vec.strings[0], 'EndNative') then
      begin
        throwError(lNum, posL, '- Expecting ''EndNative''.', mes);
        convert := txtO; exit;
      end;

      convert := txtO; exit;
    end
    else if (eic(vec.Strings[posL], 'Lbl')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting label identifier after ''Lbl''.', mes);
        convert := txtO; exit;
      end;

      txtO := txtO + UpperCase(vec.Strings[posL]);

      txtO := txtO + ': ';
 //     posL := posL + 1;
    end
    else if (eic(vec.Strings[posL], 'Goto')) then
    begin
      posL := posL + 1;

      if (posL >= vec.Count) then
      begin
        throwError(lNum, posL, '- Expecting label identifier after ''Goto''.', mes);
        convert := txtO; exit;
      end;

      txtO := txtO + 'goto ';

      txtO := txtO + UpperCase(vec.Strings[posL]);

      txtO := txtO + ' ; ';
//      posL := posL + 1;
    end
    else
    begin
      i := vec.Count;
      txtO := txtO + convertExpr(getChunk(vec, posL, i), loc, loct, 0) + ' ; ';
      posL := i;
      continue;
    end;

    posL := posL + 1;
  end;

  convert := txtO;
end;

function beginWith(strs: TStrings; str: AnsiString): Boolean;
var
  len: Longint;
  len0: Longint;
  i: Longint;
  n: Longint;
  s: AnsiString;
  rep: Boolean;
begin

beginWith := false;
for n := 0 to strs.Count - 1 do
begin
  s := strs.Strings[n];
  len := Length(s);
  len0 := Length(str);

  if (len < len0) then
  begin
    beginWith := false;
    continue;
  end;

  beginWith := true;
  rep := true;

  for i := 1 to len0 do
  begin
    if (UpperCase(s[i]) <> UpperCase(str[i])) then
    begin
      beginWith := false;
      rep := false;
      break;
    end;
  end;

  if rep then
  begin
    beginWith := true;
    exit
  end;

end;

end;

function getIndexOf(strs: TStrings; str: AnsiString): Longint;
var
  ind: Longint;
  i: Longint;
  b: Boolean;
begin

  ind := -1;

  for i := 0 to strs.Count - 1 do
  begin

    b := true; 

    if (UpperCase(strs.Strings[i]) <> UpperCase(str)) then
    begin
      b := false;
      continue;
    end;

    if b then
    begin
      ind := i;
    end;
  end;

  getIndexOf := ind;
  exit;
end;

procedure TFormBuild.btnCancelClick(Sender: TObject);
begin
  FormBuild.Close;
end;

procedure throwError(l : Longint; t : Longint; msg: AnsiString; mes: TStrings);
begin
  mes.Append('Line ' + intToStr(l) + ':' + intToStr(t) + msg);
  mes.Append('Failed to build project.');
  errorAppend := true;
end;

procedure TFormBuild.btnConvertClick(Sender: TObject);
var
  text: AnsiString;
  txtO: AnsiString;

  vec: TStrings;
  posL: Longint;
  str: AnsiString;
  i: Longint;
  n: Longint;

  argList: TStrings;
  prgFile: TStrings;
  tmpFile: TStrings;
  //prg: Boolean;
  loc: TStrings;
  typ: TStrings;
begin
  newLine := #13;

  pnlMes.Visible := true;
  btnCancel.visible := false;
  btnConvert.visible := false;
  btnClose.visible := true;

  isSaveScreen := true;
  strSize := 256;

  typList := TStringList.Create;

  esp := TStringList.Create;

  esp.Append(' ');
  esp.Append(#9);
  esp.Append(#$D);
  esp.Append(#$A);

  used := TStringList.Create;

  pList := TStringList.Create;
  pList.LoadFromFile(confDir+'procedures.lst');

  fList := TStringList.Create;
  flist.LoadFromFile(confDir+'functions.lst');

  loc := TStringList.Create;
  typ := TStringList.Create;

  ops := TStringList.Create;

  ops.LoadFromFile(confDir+'symbols.lst');

  keyw := TStringList.Create;

  keyw.LoadFromFile(confDir+'keywords.lst');

  opPar := TStringList.Create;

  opPar.LoadFromFile(confDir+'opPar.lst');

  aNumR := TStringList.Create;
  aNumC := TStringList.Create;

  clPar := TStringList.Create;

  clPar.LoadFromFile(confDir+'clPar.lst');

  consVec := TStringList.Create;
  consVec.LoadFromFile(confDir+'const.def');

  tmpFile := TStringList.Create;
  argList := TStringList.Create;

  tmpFile.LoadFromFile(confDir+'operators.def');

  opers := TStringList.Create;
  oPri := TStringList.Create;

  for i := 0 to tmpFile.Count - 1 do
  begin
    argList := getTokens(tmpFile.Strings[i]);
    for n := 0 to argList.Count - 1 do
    begin
      opers.Append(argList.Strings[n]);
      oPri.Append(intToStr(i));
    end;
  end;

  opers.Append('-neg-');
  oPri.Append('8');

  opers.Append('+pos+');
  oPri.Append('8');
  
  tmpFile.Clear;
  argList.Clear;

  errorAppend := false;

  txtMes.Clear;

  mes := txtMes.Lines;

  mes.Append('Converting source code to TIGCC project ...');

  prgName := '';

  aCount := 0;

  varList := TStringList.Create;

  varList.append('OK');
  typList.append('Variant');

  prgList := TStringList.Create;
  funcList := TStringList.Create;
  prgTxt := TStringList.Create;
  prgArgs := TStringList.Create;
  funcTxt := TStringList.Create;
  funcArgs := TStringList.Create;

  lNum := 0;
  posL := 0;

  for i := 0 to consVec.count - 1 do
  begin
    varList.Append(consVec[i]);
    typList.Append('//');
  end;


  vec := getTokens(Form1.txtCode.Lines[0]);

        if (vec.Count < 3) then
        begin
          throwError(lNum, posL, '- Expecting program declaration.', mes);
          exit;
        end;
        prgName := vec.Strings[0];
        mes.Append('Program name : ' + prgName);
        if (Length(prgName) > 8) then
        begin
          throwError(lNum, posL, '- Program name invalid.', mes);
          exit;
        end;

        posL := posL + 1;

        if (vec.Strings[posL] <> '(') then
        begin
          mes.Append('Line ' + intToStr(lNum) + ':' + intToStr(posL) + '- Expecting ''(''.');
          exit;
        end;

        posL := posL + 1;

        if (vec.Strings[posL] <> ')') then
        begin

          while (vec.Strings[posL] <> ')') do
          begin
            argList.Append(vec.Strings[posL]);

            posL := posL + 1;

            if (vec.Count <= posL) then
            begin
              mes.Append('Line ' + intToStr(lNum) + ':' + intToStr(posL) + '- Program declaration invalid.');
              exit;
            end;

            if (vec.Strings[posL] <> ',') and (vec.Strings[posL] <> ')') then
            begin
              mes.Append('Line ' + intToStr(lNum) + ':' + intToStr(posL) + '- Expecting '','' or '')''.');
              exit;
            end;

            if (vec.Strings[posL] = ')') then
            begin
              break;
            end;

            posL := posL + 1;

            if (vec.Count <= posL) then
            begin
              mes.Append('Line ' + intToStr(lNum) + ':' + intToStr(posL) + '- Program declaration invalid.');
              exit;
            end;
          end;

          if (vec.Strings[posL] <> ')') then
          begin
            mes.Append('Line ' + intToStr(lNum) + ':' + intToStr(posL) + '- Expecting '')''.');
            exit;
          end;
        end
        else if (vec.Count > 3) then
        begin
          posL := posL + 1;
          mes.Append('Line ' + intToStr(lNum) + ':' + intToStr(posL) + '- Unexpected token ''' + vec.Strings[posL] + ''' after ''(''.');
          exit;          
        end;

        str := 'Argument list : ';

        for i := 0 to argList.Count - 1 do
        begin
          loc.append(argList.Strings[i]);
          typ.Append('Variant');
          str := str + argList.Strings[i] + '; ';
        end;

        mes.Append(str);

  lNum := lNum + 1;
  posL := 0;
  vec := getTokens(Form1.txtCode.Lines[lNum]);

  if (posL >= vec.Count) or (not eic(vec.Strings[0], 'Prgm') and not eic(vec.Strings[0], 'Func')) then
  begin
    throwError(lNum, posL, '- Expecting ''Prgm'' or ''Func''.', mes);
  end;

  if eic(vec.Strings[0], 'Prgm') then
  begin
    isPrg := true;
  end
  else
  begin
    isPrg := false;
  end;

  lNum := lNum + 1;

  txtO := '';

  while (lNum < Form1.txtCode.Lines.Count) do
  begin
    vec := getTokens(Form1.txtCode.Lines[lNum]);

    posL := 0;

    if (posL < vec.Count) and (vec.Strings[0] = '''') then
    begin
      (* Comment *)
      txtO := txtO + '/* ' + Form1.txtCode.Lines[lNum] + '*/';

      lNum := lNum + 1;
      continue;
    end;

    if (posL < vec.Count) and (eic(vec.Strings[posL], 'EndPrgm') or eic(vec.Strings[posL], 'EndFunc')) then
    begin
      break;
    end;

    str := '';

    if (posL < vec.Count) and eic(vec.Strings[posL], 'Return') then
    begin
      str := str + 'freeStorage ( ) ; ';

      posL := posL + 1;

      if not isPrg then
      begin
        str := str + 'returnVariant ( ' + convertExpr(getRema(vec, posL), loc, typ, 0) + ' ) ; '
      end
      else
      begin
        str := str + 'exit ( 0 ) ; ';
      end;
    end ;

    if (str = '') then
    begin
      txtO := txtO + convert(vec, loc, typ);
    end
    else
    begin
      txtO := txtO + str;
    end;

    txtO := txtO + newLine;

    if errorAppend then
    begin
      exit;
    end;

    lNum := lNum + 1;
  end;

  txtO := toDeclaration(loc, typ, argList) + txtO;

  if (posL >= vec.Count) or (not eic(vec.Strings[posL], 'EndPrgm') and not eic(vec.Strings[posL], 'EndFunc')) then
  begin
    throwError(lNum, posL, '- Expecting ''EndPrgm'' or ''EndFunc''.', mes);
    exit;
  end;

  prgFile := TStringList.Create;

  CreateDir(txtDir.Text);
  tmpFile.LoadFromFile(confDir + 'Variant.h');

  if isUseAll then
  begin
    used.Clear;
    for i := 0 to fList.Count - 1 do
    begin
      vec := getTokens(fList[i]);
      tmpFile.Append('#define USES_' + UpperCase(vec[1]) + ' 1');
    end;
    for i := 0 to pList.Count - 1 do
    begin
      vec := getTokens(pList[i]);
      tmpFile.Append('#define USES_' + UpperCase(vec[0]) + ' 1');
    end;
  end;

  for i := 0 to used.Count - 1 do
  begin
    tmpFile.Append('#define USES_' + UpperCase(used[i]) + ' 1');
  end;

  tmpFile.SaveToFile(txtDir.Text + '\Variant.h');
  tmpFile.LoadFromFile(confDir + 'Variant.c');
  tmpFile.SaveToFile(txtDir.Text + '\Variant.c');

  tmpFile.LoadFromFile(confDir + 'Functions.c');
  tmpFile.SaveToFile(txtDir.Text + '\Functions.c');
  tmpFile.LoadFromFile(confDir + 'Procedures.c');
  tmpFile.SaveToFile(txtDir.Text + '\Procedures.c');
  tmpFile.LoadFromFile(confDir + 'Functions.h');
  tmpFile.SaveToFile(txtDir.Text + '\Functions.h');
  tmpFile.LoadFromFile(confDir + 'Procedures.h');
  tmpFile.SaveToFile(txtDir.Text + '\Procedures.h');

  prgFile.Append('#define ENABLE_ERROR_RETURN');
  if not isPrg then
  begin
    prgFile.Append('#define RETURN_VALUE //The program will return a value');
  end;

  if isSaveScreen then
  begin
    prgFile.Append('#define SAVE_SCREEN');
  end;
//  prgFile.Append('#define OPTIMIZE_ROM_CALLS');
//  prgFile.Append('#define EXECUTE_IN_GHOST_SPACE');

  prgFile.Append('#define USE_TI89              	// Compile for TI-89');
  prgFile.Append('#define USE_TI92PLUS          	// Compile for TI-92 Plus');
  prgFile.Append('#define USE_V200              	// Compile for V200');

  prgFile.Append('void ' + 'command_Stop ( ) ; ');
  prgFile.Append('void ' + 'freeStorage ( ) ; ');

  prgFile.Append('#define VARIANT_STRING_SIZE ' + intToStr(strSize));

  tmpFile.LoadFromFile(confDir + 'define.default');
  for i := 0 to tmpFile.count - 1 do
  begin
    prgFile.Append(tmpFile[i]);
  end;

  prgFile.Append('#include "Variant.h"');
  prgFile.Append('#include "Procedures.h"');
  prgFile.Append('#include "Functions.h"');

  for i := 0 to varList.count - 1 do
  begin
    prgFile.Append(typList[i] + ' ' + UpperCase(varList.Strings[i]) + ' ; ');
  end;

  for i := 0 to aCount - 1 do
  begin
    prgFile.Append ('VariantArray array_' + intToStr(i) + ' ; ');
  end;

  prgFile.Append('void ' + 'freeStorage ( ) {');

  for i := 0 to aCount - 1 do
  begin
    prgFile.Append ('freeArray ( & array_' + intToStr(i) + ' ) ; ');
  end;

  prgFile.Append('} ');

  prgFile.Append('void ' + 'command_Stop ( ) {');

  prgFile.Append('freeStorage ( ) ;');

  if not isPrg then
  begin
    prgFile.Append('returnVariant(newVariantNum(0)) ; ');
  end
  else
  begin
    prgFile.Append('exit ( 0 ) ; ');
  end;

  prgFile.Append('} ');

  for i := 0 to prgList.Count - 1 do
  begin
    str := 'void ' + UpperCase(prgList.Strings[i]) + ' (';

    vec := getTokens(UpperCase(prgArgs.Strings[i]));

    for n := 0 to vec.count - 1 do
    begin
      if (n > 0) then
      begin
        str := str + ' , ';
      end;
      str := str + 'Variant ' + vec.Strings[n];
    end;

    prgFile.Append(str + ') ; ');
  end;

  for i := 0 to funcList.Count - 1 do
  begin
    str := 'Variant ' + UpperCase(funcList.Strings[i]) + ' (';

    vec := getTokens(UpperCase(funcArgs.Strings[i]));

    for n := 0 to vec.count - 1 do
    begin
      if (n > 0) then
      begin
        str := str + ' , ';
      end;
      str := str + 'Variant ' + vec.Strings[n];
    end;

    prgFile.Append(str + ') ; ');
  end;

  for i := 0 to prgList.Count - 1 do
  begin
    str := 'void ' + UpperCase(prgList.Strings[i]) + ' (';

    vec := getTokens(UpperCase(prgArgs.Strings[i]));

    for n := 0 to vec.count - 1 do
    begin
      if (n > 0) then
      begin
        str := str + ' , ';
      end;
      str := str + 'Variant ' + vec.Strings[n];
    end;

    prgFile.Append(str + ') { ');
    prgFile.Append(prgTxt.Strings[i]);
    prgFile.Append('} ');
  end;

  for i := 0 to funcList.Count - 1 do
  begin
    str := 'Variant ' + UpperCase(funcList.Strings[i]) + ' (';

    vec := getTokens(UpperCase(funcArgs.Strings[i]));

    for n := 0 to vec.count - 1 do
    begin
      if (n > 0) then
      begin
        str := str + ' , ';
      end;
      str := str + 'Variant ' + vec.Strings[n];
    end;

    prgFile.Append(str + ') { ');
    prgFile.Append(funcTxt.Strings[i]);
    prgFile.Append('} ');
  end;

  str := '';

  for i := 0 to argList.count -1 do
  begin
    if (i > 0) then
    begin
      str := str + ', ';
    end;
    str := str + 'Variant ' + UpperCase(argList.strings[i]);
  end;

  prgFile.Append('void ' + UpperCase(prgName) + ' ( ' + str + ' ) {');
  prgFile.Append(txtO);
  prgFile.Append('}');


  prgFile.Append('void _main ( void ) { ');
  prgFile.Append('initVariantType ( ) ; ');

  for i := 0 to aCount - 1 do
  begin
    prgFile.Append ('newArray ( & array_' + intToStr(i) + ', ' + aNumR.Strings[i] + ', ' + aNumC.Strings[i] + ' ) ; ');
  end;

  for i := 0 to argList.count - 1 do
  begin
    prgFile.Append('Variant ' + UpperCase(argList.Strings[i]) + ' = fetchVariantFromArg ( ) ; '); 
  end;

//  prgFile.Append(str + ' ; ');

  str := '';

  for i := 0 to argList.count -1 do
  begin
    if (i > 0) then
    begin
      str := str + ', ';
    end;
    str := str + UpperCase(argList.Strings[i]);
  end;

  prgFile.Append(UpperCase(prgName) + ' ( ' + str + ' ) ; ');

  prgFile.Append('command_Stop ( ) ; ');

  prgFile.Append('} ');
  prgFile.Append('');

  prgFile.SaveToFile(txtDir.Text + '\Main.c');

  tmpFile.LoadFromFile(confDir + 'tigcc_project.default');

  tmpFile.Append('[Included Files]');
  tmpFile.Append('C File 1=Main.c');
  tmpFile.Append('C File 2=Variant.c');
  tmpFile.Append('C File 3=Procedures.c');
  tmpFile.Append('C File 4=Functions.c');
  tmpFile.Append('Header File 1=Variant.h');
  tmpFile.Append('Header File 2=Procedures.h');
  tmpFile.Append('Header File 3=Functions.h');

  tmpFile.SaveToFile(txtDir.Text + '\' + prgName + '.tpr');

  mes.Append('Project built successfully.');

  mes.Append(txtO);
end;

function getTokens(text: AnsiString): TStrings;
var
  mot: AnsiString;
  i: Longint;
  lv: Longint;
  posBack: Longint;
  tokens: TStrings;
  pos: Longint;
begin
  tokens := TStringList.Create;
  pos := 1;

  while (pos <= length(text)) do
  begin
    mot := '';

    while (pos <= length(text)) and (getIndexOf(esp, text[pos]) >= 0) do
    begin
      pos := pos + 1;
    end;

    (*if (pos + 1 <= length(text)) and (text[pos] = '-') and isDigit(text[pos + 1]) then
    begin

      mot := mot + text[pos];
      pos := pos + 1;

      while (pos <= length(text)) and not (getIndexOf(esp, text[pos]) >= 0) and not beginWith(ops, text[pos]) do
      begin
        mot := mot + text[pos];
        pos := pos + 1;
      end;

      if beginWith(ops, text[pos]) then
      begin
        if (mot <> '') then
        begin
          tokens.Append(mot);
        end;
        continue;
      end;
    end
    else
    begin   *)
      if beginWith(ops, text[pos]) then
      begin
        while (pos <= length(text)) and beginWith(ops, mot + text[pos]) do
        begin
          mot := mot + text[pos];
          pos := pos + 1;
        end;
        tokens.Append(mot);
        continue;
      end;
   // end;

    if (pos <= length(text)) and (text[pos] = '"') then
    begin
      mot := mot + text[pos];
      pos := pos + 1;
      while (pos <= length(text)) and ((text[pos] <> '"') or ((pos + 1 <= length(text)) and (text[pos+1] = '"'))) do
      begin
        if (pos + 1 <= length(text)) and (text[pos] = '"') and (text[pos+1] = '"') then
        begin
          mot := mot + '\';
          pos := pos + 1;
        end;

        mot := mot + text[pos];
        pos := pos + 1;
      end;

      mot := mot + '"';

    end
    else
    begin

      while (pos <= length(text)) and not (getIndexOf(esp, text[pos]) >= 0) and not beginWith(ops, text[pos]) do
      begin
        mot := mot + text[pos];
        pos := pos + 1;
      end;

      if beginWith(ops, text[pos]) then
      begin
        if (mot <> '') then
        begin
          tokens.Append(mot);
        end;
        continue;
      end;
    end;

    if (mot <> '') then
    begin
      tokens.Append(mot);
    end;

    pos := pos + 1;
  end;

  getTokens := tokens;
end;



procedure TFormBuild.btnCloseClick(Sender: TObject);
begin
  FormBuild.Close;
end;

end.
