(*
 * QUSOFT MICROSYSTMES
 * Power Basic
 * Copyright 2004 Frdric Brown
 *)

(*
 *  This file is part of Power Basic.
 *
 *  Power Basic is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.

 *  Power Basic is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with Power Basic; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit MainFrame;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Menus, ExtCtrls, ComCtrls, TabNotBk, BuildFrame, About;

type
  TForm1 = class(TForm)
    dlgSave: TSaveDialog;
    dlgOpen: TOpenDialog;
    MainMenu1: TMainMenu;
    menuFile: TMenuItem;
    itmNew: TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    Saveas1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    N2: TMenuItem;
    Convert1: TMenuItem;
    Help1: TMenuItem;
    About1: TMenuItem;
    ConverttoTIGCCproject1: TMenuItem;
    TabbedNotebook1: TTabbedNotebook;
    txtCode: TMemo;
    Panel2: TPanel;
    Panel1: TPanel;
    txtCat: TMemo;
    Panel3: TPanel;
    cmbCat: TComboBox;
    pnlFile: TPanel;
    Newfunction1: TMenuItem;
    procedure Open1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure Saveas1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure itmNewClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure txtCodeChange(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure cmbCatChange(Sender: TObject);
    procedure ConverttoTIGCCproject1Click(Sender: TObject);
    procedure Newfunction1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure updateInfos;
    procedure saveCurrentProgram;
  end;

var
  Form1: TForm1;
  currentFile : AnsiString;
  modifFile : Boolean;
  curDir: AnsiString;
  confDir: AnsiString;

implementation

{$R *.dfm}

procedure TForm1.Open1Click(Sender: TObject);
begin
  if modifFile then
  begin
    if (MessageDlg('Changes has been made to the document, do you wish to save ?', mtConfirmation, mbYesNoCancel, 0) = mrYes) then
    begin
      saveCurrentProgram;
    end;
  end;
  if dlgOpen.Execute then
  begin
    currentFile := dlgOpen.Files.Strings[0];
    txtCode.Lines.LoadFromFile(currentFile);
    modifFile := false;
    updateInfos;
    FormBuild.txtDir.Text := 'TIGCC';
  end;
end;

procedure TForm1.Exit1Click(Sender: TObject);
begin
  if modifFile then
  begin
    if (MessageDlg('Changes has been made to the document, do you wish to save ?', mtConfirmation, mbYesNoCancel, 0) = mrYes) then
    begin
      saveCurrentProgram;
    end;
  end;

  Application.Terminate;
end;

procedure TForm1.Saveas1Click(Sender: TObject);
begin
  if dlgSave.Execute then
  begin
    currentFile := dlgSave.Files.Strings[0];
    modifFile := false;
    updateInfos;
    txtCode.Lines.SaveToFile(currentFile);
  end;
end;

procedure TForm1.Save1Click(Sender: TObject);
begin
  saveCurrentProgram;
end;

procedure TForm1.itmNewClick(Sender: TObject);
var
 prgName : AnsiString;
 params : AnsiString;
begin
  prgName := InputBox('New program', 'Type in program''s name (8 characters max):', '');
  params := '';

  if (prgName = '') then
  begin
    ShowMessage('No name entered, operation canceled.');
    exit;
  end;

  if (length(prgName) > 8) then
  begin
    ShowMessage('Invalid program name, operation canceled.');
    exit;
  end
  else
  begin
    params := InputBox('New program', 'Type in parameter names, each separated by a comma (,) :', '');
  end;

  currentFile := '';
  dlgSave.FileName := prgName;
  modifFile := true;
  FormBuild.txtDir.Text := 'TIGCC';

  updateInfos;

  txtCode.Lines.Clear;

  txtCode.Lines.Append(prgName + '(' + params + ')');
  txtCode.Lines.Append('Prgm');
  txtCode.Lines.Append('');
  txtCode.Lines.Append('EndPrgm');
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  currentFile := '';
  modifFile := false;
  curDir := GetCurrentDir + '\';
  confDir:= curDir + 'conf\';
  txtCat.Lines.LoadFromFile(confDir+'Functions.cat');
end;

procedure TForm1.updateInfos();
begin
  pnlFile.Caption := 'Current file : ' + currentFile;

  if modifFile then
    pnlFile.Caption := pnlFile.Caption + ' [modified]';
end;

procedure TForm1.saveCurrentProgram ();
begin
  if currentFile <> '' then
  begin
    txtCode.Lines.SaveToFile(currentFile);
  end
  else
  begin
    if dlgSave.Execute then
    begin
      currentFile := dlgSave.Files.Strings[0];
      txtCode.Lines.SaveToFile(currentFile);
    end;
  end;
  modifFile := false;
  updateInfos;
end;

procedure TForm1.txtCodeChange(Sender: TObject);
begin
  modifFile := true;
  updateInfos;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if modifFile then
  begin
    if (MessageDlg('Changes has been made to the document, do you wish to save ?', mtConfirmation, mbYesNoCancel, 0) = mrYes) then
    begin
      saveCurrentProgram;
    end;
  end;
end;

procedure TForm1.cmbCatChange(Sender: TObject);
begin
  if (cmbCat.ItemIndex = 0) then
  begin
    txtCat.Lines.LoadFromFile(confDir+'Operators.cat');
  end
  else if (cmbCat.ItemIndex = 1) then
  begin
    txtCat.Lines.LoadFromFile(confDir+'Keywords.cat');
  end
  else if (cmbCat.ItemIndex = 2) then
  begin
    txtCat.Lines.LoadFromFile(confDir+'Functions.cat');
  end
  else if (cmbCat.ItemIndex = 3) then
  begin
    txtCat.Lines.LoadFromFile(confDir+'Constants.cat');
  end;
end;

procedure TForm1.ConverttoTIGCCproject1Click(Sender: TObject);
begin
  saveCurrentProgram();
  FormBuild.pnlMes.Visible := false;
  FormBuild.btnCancel.visible := true;
  FormBuild.btnClose.visible := false;
  FormBuild.btnConvert.visible := true;
  FormBuild.Show;
end;

procedure TForm1.Newfunction1Click(Sender: TObject);
var
 prgName : AnsiString;
 params : AnsiString;
begin
  prgName := InputBox('New function', 'Type in program''s name (8 characters max):', '');
  params := '';

  if (prgName = '') then
  begin
    ShowMessage('No name entered, operation canceled.');
    exit;
  end;

  if (length(prgName) > 8) then
  begin
    ShowMessage('Invalid function name, operation canceled.');
    exit;
  end
  else
  begin
    params := InputBox('New function', 'Type in parameter names, each separated by a comma (,) :', '');
  end;

  currentFile := '';
  dlgSave.FileName := prgName;
  modifFile := true;
  FormBuild.txtDir.Text := 'TIGCC';

  updateInfos;

  txtCode.Lines.Clear;

  txtCode.Lines.Append(prgName + '(' + params + ')');
  txtCode.Lines.Append('Func');
  txtCode.Lines.Append('');
  txtCode.Lines.Append('EndFunc');
end;

procedure TForm1.About1Click(Sender: TObject);
begin
  AboutBox.showModal();
end;

end.
